"""
Plots directly related to tenbilac objects
"""

import numpy as np
import itertools
import re
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.lines
from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib.cm
#from mpl_toolkits.axes_grid1 import make_axes_locatable

from . import err
from . import net
from . import wnet

import logging
logger = logging.getLogger(__name__)



def errevo(ax, train, showtimes=True):
	"""
	Plots the evolution of the error curve during the training, iteration per iteration.
	
	:param ax: matplotlib axes
	:param showtimes: If True, some training times are written on the curve, in minutes.
	"""

	logger.info("Preparing error evolution plot for {train}".format(train=str(train)))
	# Preparint the data:
	optiterrs_train = np.array(train.optiterrs_train)
	optiterrs_val = np.array(train.optiterrs_val)
	optits = np.arange(len(train.optitparams))
	optbatchchangeits = getattr(train, "optbatchchangeits", [])
	
	# The cpu durations
	optittimes = np.array(train.optittimes)
	cumoptittimes = np.cumsum(optittimes)/60.0 # minutes
	assert cumoptittimes.size == optittimes.size
	
	if optittimes.size > 10:
		labelstep = int(float(optittimes.size)/10.0)
		timeindices = range(labelstep, optittimes.size, labelstep)
	else:
		timeindices = []

	for optbatchchangeit in optbatchchangeits:
		ax.axvline(optbatchchangeit, color="gray")

	ax.plot(optits, optiterrs_train, ls="-", color="black", label="Training batch")
	ax.plot(optits, optiterrs_val, ls="--", color="red", label="Validation set")
	
	if showtimes:
		for i in timeindices:
			ax.annotate("{0:.1f}".format(cumoptittimes[i]), xy=(optits[i], optiterrs_val[i]), xytext=(0, 10), textcoords='offset points')
	
	ax.set_yscale('log')
	ax.set_xlabel("Iteration")
	ax.set_xlim((optits[0], optits[-1]))
	ax.set_ylabel("Cost function value ({0})".format(train.errfctname))
	ax.legend()
	ax.set_title(train.title())





def paramsevo(ax, train, wnetpart=None):
	"""
	Plots the evolution of the actual network parameters.
	
	""" 
	
	optits = np.arange(len(train.optitparams))
	
	if isinstance(train.net, wnet.WNet):
		if wnetpart == "o":
			mynet = train.net.neto
			optitparams = np.array(train.optitparams)[:,:train.net.neto.nparams()]
			
		elif wnetpart == "w":
			mynet = train.net.netw
			optitparams = np.array(train.optitparams)[:,train.net.neto.nparams():]
		else:
			raise ValueError("This is a WNet, please specify a wnetpart")
	else: # We have a normal Net:
		mynet = train.net
		optitparams = np.array(train.optitparams)
	
	paramlabels = mynet.get_paramlabels()
	multmode = False

	assert optitparams.shape[1] == mynet.nparams()
	for paramindex in range(mynet.nparams()):
		
		label = paramlabels[paramindex]
		if label.endswith("_bias"):
			ls = "--"
		elif label.endswith("_weight"):
			ls = "-"
			
		layermode = re.match("layer-(.*)-(.*)_(.*)", label).group(1)
		layername = re.match("layer-(.*)-(.*)_(.*)", label).group(2)

		if layername == "o":
			color="black"
		elif layermode == "sum":
			color="blue"
		elif layermode == "mult":
			multmode = True
			color="green"
		else:
			raise ValueError("Layer mode {} unknown".format(layermode))
		
		pla = ax.plot(optits, optitparams[:,paramindex], ls=ls, color=color)
	
	ax.set_xlabel("Iteration")
	ax.set_xlim((optits[0], optits[-1]))
	ax.set_ylabel("Network parameter value")
	
	# Now creating the legend
	
	black_patch = matplotlib.patches.Patch(color='black', label='Output layer')
	
	line = matplotlib.lines.Line2D([], [], color='black', marker='', ls="-", label='Weight')
	dashed = matplotlib.lines.Line2D([], [], color='black', marker='', ls="--", label='Bias')
	
	handles=[line, dashed, black_patch]
	
	if multmode:
		red_patch = matplotlib.patches.Patch(color='blue', label='$\Sigma$ Hidden layers')
		green_patch = matplotlib.patches.Patch(color='green', label='$\Pi$ Hidden layers')
		handles += [red_patch]
		handles += [green_patch]
	else:
		red_patch = matplotlib.patches.Patch(color='blue', label='Hidden layers')
		handles += [red_patch]

	ax.legend(handles=handles)



	
def sumevo(train, filepath=None, showtimes=True):
	"""
	Visualization of the evolution of the network parameters and error during the training,
	iteration per iteration
	
	:param showtimes: If True, some training times are written on the curve, in minutes.
	"""
	
	
	fig = plt.figure(figsize=(10, 10))
	ax = plt.subplot(2, 1, 1)
	errevo(ax, train, showtimes=showtimes)
	
	ax = plt.subplot(2, 1, 2)
	paramsevo(ax, train, wnetpart=None)
	
	#plt.tight_layout()
	if filepath is None:
		plt.show()	
	else:
		logger.info("Writing paramscurve to {0}".format(filepath))
		plt.savefig(filepath)
	plt.close() # Important, otherwise it's still around for the next plt.show()



def biasevo(train, filepath=None):
	"""
	Viz of the evolution of the individual bias terms at each iteration (massive!)
	"""
	
	nsnaps = len(train.biassnaps_it)
	if nsnaps < 1:
		logger.warning("No snapshots to biasevo plot!")
		return
		
		
	trainbiases = np.array(train.biassnaps_train) # Members of this list have shape (outputs, cases). Now its (iterations, outputs, cases)
	valbiases = np.array(train.biassnaps_val) # Members of this list have shape (outputs, cases). Now its (iterations, outputs, cases)
	
	biases = np.dstack((trainbiases, valbiases)) # shape is (iteration, output-neuron, case)
	
	assert nsnaps == biases.shape[0]
	ncas = biases.shape[2]
	
	# For colors, we get the input data
	if train.dat is None:
		logger.warning("Need dat for biasevo plot!")
		return
	meantraininputs = np.mean(train.dat.fulltraininputs, axis=0) # shape is (feature, case)
	meanvalinputs = np.mean(train.dat.valinputs, axis=0) # shape is (feature, case)
	meaninputs = np.hstack((meantraininputs, meanvalinputs)) # shape is still (feature, case)
	
	assert meaninputs.shape[1] == biases.shape[2]
	
	logger.info("Preparing biasevo plot with {} cases and {} snapshots...".format(ncas, nsnaps))
	

	fig = plt.figure(figsize=(8*train.net.no, 5*train.net.ni))
	
	for io in range(train.net.no):
		for ii in range(train.net.ni):
			ax = plt.subplot(train.net.ni, train.net.no, io*train.net.ni+1+ii)
			
			#for optbatchchangeit in optbatchchangeits:
			#	ax.axvline(optbatchchangeit, color="gray", zorder=-20)
			ax.axhline(0.0, color="gray", zorder=-20)
			
			cmap = matplotlib.cm.get_cmap("jet")
			sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=np.min(meaninputs[ii,:]), vmax=np.max(meaninputs[ii,:])))
			
			for icas in range(ncas):
				plt.plot(train.biassnaps_it, biases[:,io, icas], color=sm.to_rgba(meaninputs[ii, icas]), alpha=0.1)
		
			ax.set_xlabel("Iteration (only snapshots at minibatch-changes are shown)")
			ax.set_ylabel("Bias on target '{}'".format(train.net.onames[io]))
	
			ax.set_yscale("symlog", linthreshy=1.e-3, linscaley=1)
			
			# fake up the array of the scalar mappable. Urgh...
			#sm._A = []
			sm.set_array([])
			cax = plt.colorbar(sm)
			cax.set_label("Feature '{}' (mean across reas)".format(train.net.inames[ii]))
			
	
	plt.tight_layout()
	if filepath is None:
		plt.show()	
	else:
		logger.info("Writing biasevo to {0}".format(filepath))
		plt.savefig(filepath)
	plt.close() # Important, otherwise it's still around for the next plt.show()

	

def outdistribs(train, filepath=None):
	"""
	Viz of the output 
	"""
	ncol = 7
	fig = plt.figure(figsize=(4*ncol, 3.5*train.net.no))
	
	dat = train.dat
	net = train.net
	assert dat is not None
	
	logger.info("Computing predictions...")
	trainoutputs = np.ma.array(net.run(dat.traininputs), mask=dat.trainoutputsmask) # masked, 3D
	valoutputs =  np.ma.array(net.run(dat.valinputs), mask=dat.valoutputsmask)
	logger.info("Done")

	trainerrors = trainoutputs - dat.traintargets # 3D - 2D = 3D
	valerrors = valoutputs - dat.valtargets

	trainbiases = np.mean(trainerrors, axis=0) # 2D (node, case)
	valbiases = np.mean(valerrors, axis=0)
	
	trainstds = np.std(trainoutputs, axis=0) # 2D
	valstds = np.std(valoutputs, axis=0)
	
	valmsrbterms = err.msrb(valoutputs, dat.valtargets, rawterms=True) # 2D
	trainmsrbterms = err.msrb(trainoutputs, dat.traintargets, rawterms=True)
	
	for io in range(train.net.no):
		
		# Subplots: (lines, columns, number)
		# We collect the stuff to be plotted as 1D arrays:
		
		# Warning, ravel does ignore the mask, so we use flatten.
		
		# The simple outputs
		thiso_valoutputs = valoutputs[:,io,:].flatten().compressed()
		thiso_trainoutputs = trainoutputs[:,io,:].flatten().compressed()
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+1)
		ax.hist(thiso_valoutputs, bins=50, histtype="step", color="red", label="Validation set")
		ax.hist(thiso_trainoutputs, bins=50, histtype="step", color="black", label="Training batch")
		ax.set_yscale('log')
		ax.set_ylabel("Counts (reas and cases)")
		ax.set_xlabel("Pred. output for '{0}'".format(net.onames[io]))
		
		# The targets
		thiso_valtargets = dat.valtargets[io,:].flatten()
		thiso_traintargets = dat.traintargets[io,:].flatten()
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+2)
		ax.hist(thiso_valtargets, bins=20, histtype="step", color="red", label="Validation set")
		ax.hist(thiso_traintargets, bins=20, histtype="step", color="black", label="Training batch")
		ax.set_ylabel("Counts (cases)")
		ax.set_xlabel("Targets '{0}'".format(net.onames[io]))

		# The prediction errors
		thiso_valerrors = valerrors[:,io,:].flatten().compressed()
		thiso_trainerrors = trainerrors[:,io,:].flatten().compressed()
		histrange = (-1.5, 1.5)
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+3)
		ax.hist(thiso_valerrors, bins=50, histtype="step", color="red", label="Validation set", range=histrange)
		ax.hist(thiso_trainerrors, bins=50, histtype="step", color="black", label="Training batch", range=histrange)
		ax.set_ylabel("Counts (reas and cases)")
		ax.set_xlabel("Errors of pred. '{0}'".format(net.onames[io]))
		ax.set_yscale('log')
		ax.set_xlim(histrange)

		# The biases
		thiso_valbiases = valbiases[io,:].flatten().compressed()
		thiso_trainbiases = trainbiases[io,:].flatten().compressed()
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+4)
		ax.hist(thiso_valbiases, bins=50, histtype="step", color="red", label="Validation set")
		ax.hist(thiso_trainbiases, bins=50, histtype="step", color="black", label="Training batch")	
		ax.set_yscale('log')	
		ax.set_ylabel("Counts (cases)")
		ax.set_xlabel("Bias of pred. for '{0}'".format(net.onames[io]))

		# The stds
		thiso_valstds = valstds[io,:].flatten().compressed()
		thiso_trainstds = trainstds[io,:].flatten().compressed()
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+5)
		ax.hist(thiso_valstds, bins=50, histtype="step", color="red", label="Validation set")
		ax.hist(thiso_trainstds, bins=50, histtype="step", color="black", label="Training batch")	
		ax.set_yscale('log')	
		ax.set_ylabel("Counts (cases)")
		ax.set_xlabel("STD of pred. for '{0}'".format(net.onames[io]))


		# Against each other
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+6)
		assert thiso_trainstds.size == thiso_trainbiases.size # Indeed even if they were masked, their masks should be identical
		assert thiso_valstds.size == thiso_valbiases.size
		ax.plot(thiso_valbiases, thiso_valstds, marker=".", ms=2, ls="None", color="red", label="Validation set", rasterized=True)
		ax.plot(thiso_trainbiases, thiso_trainstds, marker=".", ms=2, ls="None", color="black", label="Training batch", rasterized=True)	
		ax.set_ylabel("STD of pred. for '{0}'".format(net.onames[io]))
		ax.set_xlabel("Bias of pred. for '{0}'".format(net.onames[io]))

		
		# The MSRB terms:
		thiso_valmsrbterms = valmsrbterms[io,:].flatten().compressed()
		thiso_trainmsrbterms = trainmsrbterms[io,:].flatten().compressed()
		ax = plt.subplot(train.net.no, ncol, (io*ncol)+7)
		ax.hist(thiso_valmsrbterms, bins=50, histtype="step", color="red", label="Validation set")
		ax.hist(thiso_trainmsrbterms, bins=50, histtype="step", color="black", label="Training batch")
		ax.set_ylabel("Counts (cases)")
		ax.set_xlabel("Relative biases of '{0}'".format(net.onames[io]))
		try:
			ax.set_yscale('log')
		except:
			logger.warning("Probably you have only one rea, this plot failed.")
			plt.close()
			return()
		
		
	plt.tight_layout()
	if filepath is None:
		plt.show()	
	else:
		logger.info("Writing outdistribs to {0}".format(filepath))
		plt.savefig(filepath)
	plt.close() # Important, otherwise it's still around for the next plt.show()

	
	
def errorinputs(train, filepath=None, io=0):
	"""
	Viz of the prediction errors as function of inputs
	
	:param io: the index of the ouput I should use. If you have only one neuron, this is 0.
	
	"""
	nlines = 3
	
	fig = plt.figure(figsize=(3.3*train.net.ni, 3.5*nlines))
	plt.figtext(0.5, 1.0, train.title(), ha="center", va="top")
	
	dat = train.dat
	net = train.net
	assert dat is not None
	
	logger.info("Computing predictions...")
	trainoutputs = np.ma.array(net.run(dat.traininputs), mask=dat.trainoutputsmask)
	valoutputs =  np.ma.array(net.run(dat.valinputs), mask=dat.valoutputsmask)
	logger.info("Done")
	
	trainerrors = trainoutputs - dat.traintargets # 3D - 2D = 3D: (rea, label, case)
	valerrors = valoutputs - dat.valtargets # idem
	
	#valmsrbterms = err.msrb(valoutputs, dat.valtargets, rawterms=True)
	#trainmsrbterms = err.msrb(trainoutputs, dat.traintargets, rawterms=True)
	
		
	for ii in range(train.net.ni):
			
			
			
		ax = plt.subplot(nlines, train.net.ni, ii+1)
			
		ax.hist(np.ravel(dat.valinputs[:,ii,:]), bins=50, histtype="step", color="red", label="Validation set")
		ax.hist(np.ravel(dat.traininputs[:,ii,:]), bins=50, histtype="step", color="black", label="Training batch")
		ax.set_yscale('log')
		if ii == 0:
			ax.set_ylabel("Counts (reas + cases, masked = 0)")
		#if ii == int(train.net.ni/2):
		#	ax.set_title(train.title())
		#ax.set_xlabel("Input '{0}'".format(net.inames[ii]))
		ax.set_xticklabels([]) # Hide x tick labels
		ax.set_xlim(-1.0, 1.0)
		# in this plot the masked inputs are shown with a value of 0, this is desired and expected.
	
		
		# Now we viz the biases and stds of the predictions (along the reas)
		
		# The biases and stds on the validation set:
		valbiases = np.mean(valerrors, axis=0)[io,:] # 1D (case)
		valstds = np.std(valoutputs, axis=0)[io,:]
		valbiases = np.tile(valbiases, (dat.valinputs.shape[0], 1)) # Inflated to 2D (rea, case), with all reas having the same value.
		valstds = np.tile(valstds, (dat.valinputs.shape[0], 1))
		
		
		# We want to get a version of these particular inputs with a mask, selecting only one node
		valinputs = dat.valinputs[:,ii:ii+1,:] # 3D : (rea, 1, case)
		valinputs = np.ma.array(valinputs, mask=dat.valoutputsmask) # Here an input is masked if any other input of that rea was masked.
		assert valinputs.shape[1] == 1
		valinputs = valinputs[:,0,:] # 2D : (rea, case)
		
		assert valbiases.shape == valinputs.shape
		assert valstds.shape == valinputs.shape
		

		# For the plot, we only use points for which the inputs are unmasked 
		valbiases = valbiases.flatten()
		valstds = valstds.flatten()
		valinputs = valinputs.flatten()
		valbiases = np.ma.array(valbiases, mask=valinputs.mask) # copying the mask
		valstds = np.ma.array(valstds, mask=valinputs.mask)
		valbiases = valbiases.compressed() # Getting rid of the masked elements
		valstds = valstds.compressed()
		valinputs = valinputs.compressed()
		
		assert valinputs.size == valbiases.size
		assert valinputs.size == valstds.size
		
		
		ax = plt.subplot(nlines, train.net.ni, train.net.ni+ii+1)
		ax.plot(valinputs, valbiases, marker=".", color="red", ls="None", ms=1)
		ax.axhline(0.0, color="black", lw=1, ls="--")
		if ii == 0:
			ax.set_ylabel("Bias of pred. for '{0}'".format(net.onames[io]))
		else:
			ax.set_yticklabels([]) # Hide y tick labels
		#ax.set_xlabel("Input '{0}'".format(net.inames[ii]))
		ax.set_xlim(-1.0, 1.0)
		ax.set_xticklabels([]) # Hide x tick labels
		
		ax = plt.subplot(nlines, train.net.ni, 2*train.net.ni+ii+1)
		ax.plot(valinputs, valstds, marker=".", color="red", ls="None", ms=1)
		if ii == 0:
			ax.set_ylabel("STD of pred. for '{0}'".format(net.onames[io]))
		else:
			ax.set_yticklabels([]) # Hide y tick labels
		ax.set_xlabel("Input '{0}'".format(net.inames[ii]))
		ax.set_xlim(-1.0, 1.0)
		
		
	plt.tight_layout()
	if filepath is None:
		plt.show()	
	else:
		logger.info("Writing errorinputs to {0}".format(filepath))
		plt.savefig(filepath)
	plt.close() # Important, otherwise it's still around for the next plt.show()

def draw_link(ax, start, end, **kwargs):
	"""
	Computes a Bezier curve between two points (`start` and `end`) on an axis `ax`
	
	:param ax: the axis to draw on
	:param start: The starting point, must be an array [Sx, Sy]
	:param end: The end point of the curve, must be an array [Ex, Ey]
	
	Any additional kawrgs are directly passed to `patches.PathPatch` to control the line style.
	"""
	
	verts = [
			(start[0], start[1]),  # P0
			(start[0]+0.5, start[1]), # P1
			(end[0]-0.5, end[1]), # P2
			(end[0], end[1]), # P3
    ]

	codes = [Path.MOVETO,
         Path.CURVE4,
         Path.CURVE4,
         Path.CURVE4,
         ]
	
	path = Path(verts, codes)
	patch = patches.PathPatch(path, facecolor='None', **kwargs)
	ax.add_patch(patch)
	
def scale_bias(b, scale=75.):
	"""
	Returns the scaled biased for the `plot.netviz` plot.
	
	:param b: the bias to scale
	:param scale: the scale parameter to be applied on the absolute value of `b`. Default=20
	"""
	return scale * np.abs(b)
	
def scale_weight(w, scale=2.):
	"""
	Returns the scaled weight for the `plot.netviz` plot.
	
	:param w: the weight to scale
	:param scale: the scale parameter to be applied on the absolute value of `w`. Default=2
	"""
	return scale * np.abs(w)

def get_color(v, pos="orange", neg="navy"):
	"""
	Returns a color code for the scalar value `v`.
	
	:param v: scalar value
	:param pos: name or color-code for the positive values
	:param neg: name or color-code for the negative values
	"""
	if v > 0:
		c = pos
	else:
		c = neg
	return c

def get_symbol(mode, latex=True):
	"""
	Returns the symbol used for displaying the layer mode
	"""
	if latex:
		if mode == "sum":
			return r"$\sum$"
		elif mode == "mult":
			return r"$\prod$"
		else:
			raise NotImplemented("mode not recognised")
	else:
		if mode == "sum":
			return "\Sigma"
		elif mode == "mult":
			return "\Pi"
		else:
			raise NotImplemented("mode not recognised")


def netviz(net, title="", legend=True, filepath=None):
	"""
	Draws a visualisation of the network in the style of the `Tensorflow` playground.
	
	:param net: the network
	:param title: The string to display as title.
	:param legend: Whether to show the legend for the tickness of the lines and points.
	:param filepath: The file path to save the data to. If `None` (default) shows the figures.
	"""
	
	nmax = np.amax([net.ni, net.no, np.amax(net.nhs)]) * 1.
	
	fig = plt.figure()
	ax = fig.add_subplot(111)
	
	write_mode = False
	for l in net.layers:
		if not l.mode == "sum":
			write_mode = True
	
	biasm = "^"

	plt.title(title)
	
	plt_kwargs = {'marker':'s', 's':35, 'c':'k', 'zorder':1}

	for ii, l in enumerate(net.layers):
		dy = nmax / 2 - (l.ni * 1.) / 2
		lnis = np.arange(l.ni) + dy
		
		plt.scatter(np.zeros_like(lnis) + ii, lnis, **plt_kwargs)
		
		if ii == 0 :
			# Write the name of the features
			inames = [r"$\mathrm{{{}}}$".format(i) for i in net.inames]
			for iii, inp in enumerate(inames):
				plt.text(ii-0.08, iii+dy, inp.replace("_", "\_"), horizontalalignment='right', verticalalignment='center')
		
		if ii >= len(net.layers) - 1: 
			flnis = np.arange(net.no)
			dyf = (nmax - (net.no * 1.)) * 0.5
		else:
			flnis = np.arange(net.layers[ii+1].ni)
			dyf = (nmax - (net.layers[ii+1].ni * 1.)) * 0.5
		
		for iw, w in enumerate(l.weights):
			# Draw the weights
			for il, link in enumerate(w):
				draw_link(ax, start=[ii, lnis[il]], 
						end=[ii+1, flnis[iw]+dyf], lw=scale_weight(link), edgecolor=get_color(link), zorder=-1)
			
			# Draw the biases	
			plt.scatter([ii+1.], [flnis[iw]+dyf+0.12], c=get_color(l.biases[iw]), edgecolors="None", s=scale_bias(l.biases[iw]), marker=biasm)
	
		if write_mode:
			plt.text(ii+1., flnis[0]+dyf-0.25, get_symbol(l.mode), horizontalalignment='center', size=7)
	
	# Draw output
	nos = np.arange(net.no)
	plt.scatter(np.zeros_like(nos) + ii + 1, nos+dyf, **plt_kwargs)
	
	# Name the output
	onames = [r"$\mathrm{{{}}}$".format(i) for i in net.onames]
	for iii, inp in enumerate(onames):
		plt.annotate(inp.replace("_", "\_"), xy=(ii+1.08, iii+dyf), horizontalalignment='left', verticalalignment='center')
		
	# Draw legend
	if legend:
		ws = [-1.,-0.5,0.5,1.]
		for iw, w in enumerate(ws):
			yy = nos[-1]+dyf - 1. - iw * 0.2
			plt.annotate(r"$%1.1f$" % w, xy=(ii+0.95, yy), horizontalalignment='right', verticalalignment='center')
			plt.scatter([ii+1], [yy], c=get_color(w), edgecolors="None", s=scale_bias(w), marker=biasm)
			plt.plot([ii+1.07, ii+1.47], [yy, yy], c=get_color(w), lw=scale_weight(w))
	
	# Taking care of a few things
	plt.xlim([-0.23 * (len(net.nhs) + 2),ii+1.7])
	plt.tight_layout()
	plt.axis('off')
	plt.gca().invert_yaxis()
	
	if filepath is None:
		plt.show()	
	else:
		logger.info("Writing netviz to {0}".format(filepath))
		plt.savefig(filepath)
	plt.close() # Important, otherwise it's still around for the next plt.show()




def summaryerrevo(committee, filepath=None, ax=None):
	"""
	First take at plotting the error curves of committee members, to compare their performances.
	
	:param committee: a list of Training objects
	"""

	if ax is None:
		fig = plt.figure(figsize=(14, 10))
		ax = plt.subplot(1, 1, 1)
	else:
		fig = None
	
	# We sort the committee:
	committee = sorted(committee, key=lambda trainobj: trainobj.optiterrs_train[-1], reverse=True)
	
	logger.info("Preparing summary plot with {} members...".format(len(committee)))
	coloriter=iter(plt.cm.plasma_r(np.linspace(0,1,len(committee))))
		
	for trainobj in committee:

		# Preparint the data:
		optiterrs_train = np.array(trainobj.optiterrs_train)
		optiterrs_val = np.array(trainobj.optiterrs_val)
		optits = np.arange(len(trainobj.optitparams))
		optbatchchangeits = getattr(trainobj, "optbatchchangeits", [])
		
		trainerr = trainobj.optiterrs_train[-1]
		valerr = trainobj.optiterrs_val[-1]
		valerrratio = valerr / trainerr
		
		color = next(coloriter)
		ax.plot(optits, optiterrs_train, ls="-", color=color, label="'{}': {:.2e} ({:.1f})".format(
			trainobj.name, trainerr, valerrratio))
		ax.plot(optits, optiterrs_val, ls=":", color=color) # No label for the validation

	
	if fig is not None:
		ax.set_yscale('log')
		#ax.set_xscale('log')
		ax.set_xlabel("Iteration")
		#ax.set_xlim((optits[0], optits[-1]))
		ax.set_ylabel("Cost function value")
		ax.legend()
		#ax.set_title(train.title())


	if fig is not None:
		plt.tight_layout()
		if filepath is None:
			plt.show()	
		else:
			logger.info("Writing summaryerrevo to {}".format(filepath))
			plt.savefig(filepath)
		plt.close() # Important, otherwise it's still around for the next plt.show()






#def checkdata(data, filepath=None):
#	"""
#	Simply plots histograms of the different features
#	Checks normalization
#	
#	:param data: 2D or 3D numpy array with (feature, case) or (rea, feature, case).
#	:type data: numpy array
#	
#	"""
#
#	fig = plt.figure(figsize=(10, 10))
#
#	if data.ndim == 3:
#
#		for i in range(data.shape[1]):
#			
#			ax = fig.add_subplot(3, 3, i)
#			vals = data[:,i,:].flatten()
#			ran = (np.min(vals), np.max(vals))
#			
#			ax.hist(vals, bins=100, range=ran, color="gray")
#			#ax.set_xlabel(r"$\theta$ $\mathrm{and}$ $\hat{\theta}$", fontsize=18)
#			#ax.set_ylabel(r"$d$", fontsize=18)
#			#ax.set_xlim(-1.2, 2.4)
#			#ax.set_ylim(1.6, 3.1)
#
#	plt.tight_layout()
#	plt.show()
#
#def errorcurve(train, filepath=None):
#	"""
#	Simple plot of the error curve generated during the training of a network
#	
#	"""
#	
#	fig = plt.figure(figsize=(10, 10))
#
#	
#	# The cost function calls:
#	opterrs = np.array(train.opterrs)
#	optcalls = np.arange(len(train.opterrs)) + 1
#	plt.plot(optcalls, opterrs, "r-")
#	
#	# The "iterations":
#	optiterrs = np.array(train.optiterrs)
#	optitcalls = np.array(train.optitcalls)
#	optits = np.arange(len(train.optiterrs)) + 1
#	
#	plt.plot(optitcalls, optiterrs, "b.")
#	ax = plt.gca()
#	for (optit, optiterr, optitcall) in zip(optits, optiterrs, optitcalls):
#		if optit % 5 == 0:
#		    ax.annotate("{0}".format(optit), xy=(optitcall, optiterr))
#
#
#	ax.set_yscale('log')
#	plt.xlabel("Cost function call")
#	plt.ylabel("Cost function value")
#
#	plt.tight_layout()
#	plt.show()	



import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)



for subfield in config.great3.subfields:

	logger.info("Working on subfield {}".format(subfield))

	catpath = config.great3.subpath(subfield, "simmeas", config.datasets["train-weight"], "groupmeascat.pkl")
	cat = momentsml.tools.io.readpickle(catpath)
	#print momentsml.tools.table.info(cat)
	#exit()
	
	traindir = config.great3.subpath(subfield, "ml", config.datasets["train-shear"])
	predcat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist, traindir)

	predcatpath = config.great3.subpath(subfield, "simmeas", config.datasets["train-weight"], "groupmeascat_predforw.pkl")
	momentsml.tools.io.writepickle(predcat, predcatpath)
	
	



"""
The entry point to Tenbilac: functions defined here take care of running committees
(i.e., ensembles of networks) and communicating with config files. Hence the name "com",
for communication, committees, common.

Its mission is to replace the messy tenbilacwrapper of MegaLUT.

Tenbilac is a class, but is NOT designed to be "kept" (in a pickle) between setup, training and predicting.
All the info is in the config files, there are NO secret instance attributes worth of keeping.

"""

from ConfigParser import SafeConfigParser
import multiprocessing
import os
import shutil
import glob
import datetime
import numpy as np

import logging
logger = logging.getLogger(__name__)

from . import train
from . import utils
from . import data
from . import net
from . import multnet
from . import train
from . import opt
from . import parmap
from . import plot


class Tenbilac():
	
	
	def __init__(self, configpath, configlist=None):
		"""Constructor, does not take a ton of arguments, just a path to a config file.
		Alternatively, the configpath can also just point to a directory (typically an existing tenbilac workdir),
		and the alphabetically *last* .cfg file in this directory will be used, AND this configpath will also be used as workdir,
		no matter what the config file says.
		
		:param configlist: An optional list of settings that will have precedence over what is
			written in the configfile. The structure is a list of tuples containing 3 strings (Section, Option, Value), such as
			[("setup", "workdir", "bla")]
		"""
		
		self.configpath = configpath
		self.config = SafeConfigParser(allow_no_value=True)
		
		
		if os.path.isfile(configpath): # Then we just read it
			logger.info("Reading in config from {}...".format(configpath))
			self.config.read(configpath)
		
		elif os.path.isdir(configpath): # We read the alphabetically last config file in this directory
			cfgfilepaths = sorted(glob.glob(os.path.join(configpath, "*.cfg")))
			if len(cfgfilepaths) == 0:
				raise RuntimeError("No config file found in {}!".format(self.configpath))
			
			cfgfilepath = cfgfilepaths[-1]
			logger.info("Found {} config files in dir, reading in {}...".format(len(cfgfilepaths), cfgfilepath))
			self.config.read(cfgfilepath)
			self.workdir = configpath
		
		else:
			raise RuntimeError("File or dir {} does not exist!".format(self.configpath))
				
		if configlist:
			logger.info("Using additional options: {}".format(configlist))
			for param in configlist:
				self.config.set(param[0], param[1], param[2])
				
		# If the name is not specified in the config, we use:
		# - the filename of the config file if configpath is a file
		# - the directory name, if configpath is a directory
		self.name = self.config.get("setup", "name") 
		if self.name is None or len(self.name.strip()) == 0: # if the ":" is missing as well, confirparser reads None
			# Then we use the filename
			self.name = os.path.splitext(os.path.basename(configpath))[0]

		# The workdir to be used
		if os.path.isdir(configpath):
			logger.info("Setting workdir to configpath")
			self.workdir = configpath
		else:
			self.workdir = self.config.get("setup", "workdir")
		
		# For easy access, we point to a few configuration items:
		self.inputnormerpath = os.path.join(self.workdir, "input_normer.pkl")
		self.targetnormerpath = os.path.join(self.workdir, "target_normer.pkl")
			
		logger.info("Constructed Tenbilac '{self.name}' with workdir '{self.workdir}'.".format(self=self))
		
		# Now we take care of the logging, in a brutal way. The methods will redicrec all tenbilac log to a file.
		# The code below only sets the logging up. It has to be activated and deactivated inside the methods.
		# Did not yet find the optimal way of doing this (decorator as class method ?)
		if self.config.getboolean("setup", "logtofile"):
			self.logger = logging.getLogger("tenbilac")
			logpath = os.path.join(self.workdir, "log.txt")
			logger.info("Tenbilac is set to log to {}".format(logpath))
			self.logfilehandler = logging.FileHandler(logpath, delay=True)
			self.logfilehandler.setLevel(logging.DEBUG)
			self.logfilehandler.setFormatter(logging.Formatter("PID %(process)d: %(levelname)s: %(name)s(%(funcName)s): %(message)s"))
			
			# Even if the file is only openend at the first message (delay=True in FileHandler above),
			# we do have to make sure the directories exist, so that logging can be done at any time.
			if not os.path.exists(self.workdir):
				os.makedirs(self.workdir)

	def _activatefilelog(self):
		"""
		Activate the logging of all tenbilac to a file.
		Note that only the method self.trian uses this so far.
		"""
		if self.config.getboolean("setup", "logtofile"):
			self.logger.addHandler(self.logfilehandler)
			self.logger.propagate = False
	
	def _deactivatefilelog(self):
		"""
		The problem: if a child method would call this, the logging-to-file would stop also for the parent calling the child.
		"""
		if self.config.getboolean("setup", "logtofile"):
			self.logger.removeHandler(self.logfilehandler)
			self.logger.propagate = True

	def __str__(self):
		return "Tenbilac '{self.name}'".format(self=self)
			

	def train(self, inputs, targets, inputnames=None, targetnames=None, auxinputs=None):
		"""
		Make and save normers
		Norm data with new normers
		Prepares training objects. TODO: If some exist, takesover their states
		Runs all thoses trainings with multiprocessing
		Analyses and compares the results obtained by the different members
		
		
		About multiprocessing: 
		One could split this task as early as possible and do all the construction of the Training and Net objects already in a pool.
		However, this preparation is very fast, and so for ease of debugging we keep this in normal loop.
		Also, we avoid duplicating the input data, which might be MASSIVE. We also avoid writing this input data to disk, and having
		the independent processes reading it again.
		
		"""
		
		# Can be used to save files at the level of this wrapper.
		startdt = datetime.datetime.now()
		
		self._activatefilelog()
		
		# For this wrapper, we only allow 3D inputs and 2D targets.
		if (inputs.ndim) != 3 or (targets.ndim) != 2:
			raise ValueError("This wrapper only accepts 3D inputs and 2D targets, you have {} and {}".format(inputs.shape, targets.shape))
		
		# Just for logging
		if os.path.exists(self.workdir):
			logger.info("The workdir already exists.")
		else:
			logger.info("The workdir does not exist, it will be created.")
		
		
		# Creating the normers and norming, if desired
		writeinputnormer = False # Should the new normers be written to disk, later?
		writetargetnormer = False
		if self.config.getboolean("norm", "oninputs"):
			logger.info("{}: normalizing training inputs...".format((str(self))))
			if self.config.getboolean("norm", "takeover") and os.path.exists(self.inputnormerpath):
				self.input_normer = utils.readpickle(self.inputnormerpath)
			else: # We make a new one:
				writeinputnormer = True
				self.input_normer = data.Normer(inputs, type=self.config.get("norm", "inputtype"))
			# And we norm the data:
			inputs = self.input_normer(inputs)
		else:
			logger.info("{}: inputs do NOT get normed.".format((str(self))))
		
		if self.config.getboolean("norm", "ontargets"):
			logger.info("{}: normalizing training targets...".format((str(self))))
			if self.config.getboolean("norm", "takeover") and os.path.exists(self.targetnormerpath):
				self.target_normer = utils.readpickle(self.targetnormerpath)
			else:
				writetargetnormer = True
				self.target_normer = data.Normer(targets, type=self.config.get("norm", "targettype"))
			targets = self.target_normer(targets)
		else:
			logger.info("{}: targets do NOT get normed.".format((str(self))))
		
		
		# And grouping them into a Traindata object:
		self.traindata = data.Traindata(
			inputs, targets, auxinputs=auxinputs,
			valfrac=self.config.getfloat("train", "valfrac"),
			shuffle=self.config.getboolean("train", "shuffle")
			)
		
		# Setting up the network- and training-objects according to the config
		nmembers = self.config.getint("net", "nmembers")
		self.committee = [] # Will be a list of Training objects.
		
		ni = inputs.shape[1]
		no = targets.shape[0]
		nettype = self.config.get("net", "type")
		
		logger.info("{}: Building a committee of {}s with {} members...".format(str(self), nettype, nmembers))
		
		for i in range(nmembers):
		
			# We first create the network
			if nettype == "Net":
				netobj = net.Net(
					ni=ni,
					nhs=list(eval(self.config.get("net", "nhs"))),
					no=no,
					actfctname=self.config.get("net", "actfctname"),
					oactfctname=self.config.get("net", "oactfctname"),
					multactfctname=self.config.get("net", "multactfctname"),
					inames=inputnames,
					onames=targetnames,
					name='{}_{}'.format(self.name, i)
					)
			elif nettype == "MultNet":
				netobj = multnet.MultNet(
					ni=ni,
					nhs=list(eval(self.config.get("net", "nhs"))),
					mwlist=list(eval(self.config.get("net", "mwlist"))),
					no=no,
					actfctname=self.config.get("net", "actfctname"),
					oactfctname=self.config.get("net", "oactfctname"),
					multactfctname=self.config.get("net", "multactfctname"),
					inames=inputnames,
					onames=targetnames,
					name='{}_{}'.format(self.name, i)
					)
			else:
				raise RuntimeError("Don't know network type '{}'".format(nettype))
			
			
			# A directory where the training can store its stuff
			trainobjdir = os.path.join(self.workdir, "member_{}".format(i))
			trainobjpath = os.path.join(trainobjdir, "Training.pkl")
			plotdirpath = os.path.join(trainobjdir, "plots")
			
			# Let's see if an existing training is available in these directories:
			oldtrainobj = None
			if self.config.getboolean("train", "takeover") and os.path.exists(trainobjpath):
				# Then we read the existing training (before the new training has any chance to write files...)
				# (The acutal takeover will happen later).
				logger.info("Reading in existing training... ")
				oldtrainobj = utils.readpickle(trainobjpath)
			
			# Now we create the Training object, with the new network and the traindata
			
			if self.config.getboolean("train", "useregul"):
				raise RuntimeError("Regul wrapper not implemented, code has to be cleaned first.")
			
			trainobj = train.Training(	
				netobj,
				self.traindata,
				errfctname=self.config.get("train", "errfctname"),
				regulweight=None,
				regulfctname=None,
				itersavepath=trainobjpath,
				saveeachit=self.config.getboolean("train", "saveeachit"),
				autoplotdirpath=plotdirpath,
				autoplot=self.config.getboolean("train", "autoplot"),
				logpath=None, # Does not work well with multi-cpu, unfortunately.
				trackbiases=self.config.getboolean("train", "trackbiases"),
				verbose=self.config.getboolean("train", "verbose"),
				name='{}_{}'.format(self.name, i)
				)
			
			# We keep the directories at hand with this object
			trainobj.dirstomake=[trainobjdir, plotdirpath]
			
			# If desired, we take over a previous training.
			# If we don't take over, we have to start from identity and add noise
			
			if oldtrainobj is None:
			
				if self.config.getboolean("net", "startidentity"):
					trainobj.net.setidentity(
						onlyn=self.config.getint("net", "onlynidentity")
						)
					if nettype == "MultNet": # We have to call multini again!
						trainobj.net.multini()
					
				trainobj.net.addnoise(
					wscale=self.config.getfloat("net", "ininoisewscale"),
					bscale=self.config.getfloat("net", "ininoisebscale"),
					multwscale=self.config.getfloat("net", "ininoisemultwscale"),
					multbscale=self.config.getfloat("net", "ininoisemultbscale")
					)
			
			else: # We do take over
				logger.info("Taking over an existing training...")
				trainobj.takeover(oldtrainobj)

				
			
			# We add this new Training object to the committee
			self.committee.append(trainobj)
			
		assert len(self.committee) == nmembers
		
		
		# Creating the directories
		if not os.path.isdir(self.workdir):
			os.makedirs(self.workdir)
		for trainobj in self.committee:
			for dirpath in trainobj.dirstomake:
				if not os.path.isdir(dirpath):
					os.makedirs(dirpath)

		# Writing the config into the workdir, with a timestamp in the filename
		if self.config.getboolean("setup", "copyconfig"):
			configcopyname = self.name + "_" + datetimestr(startdt) + ".cfg"
			configcopypath = os.path.join(self.workdir, configcopyname)
			with open(configcopypath + "_running", 'wb') as configfile: # For now, we add this "_running". Will be removed when done.
				self.config.write(configfile)
			# No, we don't copy the file as (1) it might already have changed and (2) some options might have been passed as configlist.
			#shutil.copy(self.configpath, configcopypath + "_running")


		# Saving the normers, now that we have the directories
		if writeinputnormer:
			utils.writepickle(self.input_normer, self.inputnormerpath)
		if writetargetnormer:
			utils.writepickle(self.target_normer, self.targetnormerpath)
	
		# Preparing the training configuration. So far, we train the committee with identical params.
		# In future, we could do this differently.
		
		# Get the name of the algorithm to use:
		algo = self.config.get("train", "algo")
		# Now we get the associated section as a dict, and "map" eval() on this dict to get ints as ints, floats as floats, bools as bools...
		trainkwargdict = {k: eval(v) for (k, v) in self.config.items("algo_" + algo)}
		
		
		# We add some further training parameters to this dict:
		trainkwargdict["algo"] = algo
		trainkwargdict["mbsize"] = None
		trainkwargdict["mbfrac"] = self.config.getfloat("train", "mbfrac")
		trainkwargdict["mbloops"] = self.config.getint("train", "mbloops")
		
		
		# And attach this dict to the committee members, to keep the map() very simple.
		for trainobj in self.committee:
			trainobj._trainkwargdict = trainkwargdict
		
		
		# We are ready to start the training (optimization)
		
		ncpu = self.config.getint("train", "ncpu")
		
		if ncpu == 0:
			try:
				ncpu = multiprocessing.cpu_count()
			except:
				logger.warning("multiprocessing.cpu_count() is not implemented!")
				ncpu = 1
			
		logger.info("{}: Starting the training on {} CPUs".format(str(self), ncpu))
		
		if ncpu == 1:
			# The single-processing version (not using multiprocessing to keep it easier to debug):
			logger.debug("Not using multiprocessing")
			map(_trainworker, self.committee)

		else:
			# multiprocessing map: # Does not work
			#pool = multiprocessing.Pool(processes=ncpu)
			#pool.map(_trainworker, self.committee)
			#pool.close()
			#pool.join()
			
			parmap.parmap(_trainworker, self.committee, ncpu)
		
		logger.info("{}: done with the training".format(str(self)))
		
		if self.config.getboolean("setup", "copyconfig"):
			# We remove the "_running"
			if os.path.exists(configcopypath + "_running"):
				os.rename(configcopypath + "_running", configcopypath) # We take care reusign the copy, and not copying again what might already have changed...
		
		if self.config.getboolean("setup", "minimize"):
			self.minimize(dt=startdt)
					
		# We close with a summary of the results
		self.summary()
	
		self._deactivatefilelog()

	def _readmembers(self):
		"""
		A method that finds available committee members by itself, exploring the file system, and reads them in.
		"""	
		
		trainpaths = sorted(glob.glob(os.path.join(self.workdir, "member_*/Training.pkl")))
		logger.info("Found {} committee members to read in...".format(len(trainpaths)))
		self.committee = [utils.readpickle(trainpath) for trainpath in trainpaths]
		return trainpaths # Potentially useful
	
	
	def summary(self):
		"""
		Summarizes the training performance of committee members.
		"""
		
		self._readmembers()
		
		# First we just write some log info, to demonstrate the idea:
		trainlistsummary(self.committee)
		
		# And we create a plot
		if self.config.getboolean("train", "autoplot"):
			plotsdirpath = os.path.join(self.workdir, "plots")
			if not os.path.exists(plotsdirpath):
				os.makedirs(plotsdirpath)
			
			plot.summaryerrevo(self.committee, filepath=os.path.join(plotsdirpath, "summaryerrevo.png"))
		

	
	def minimize(self, destdir=None, dt=None):
		"""
		Prepares a self-sufficient "copy" of the current workdir containing all the information needed for predictions, but
		without all the (potentially large) logs, plots, etc.
		
		:params destdir: path to where to save the results
		:params dt: if destdir is not given, datetime object to use to name the destdir
		
		"""
		
		if dt is None:
			dt = datetime.datetime.now()
		if destdir is None:
			destdir = os.path.join(self.workdir, "mini_" + self.name + "_" + datetimestr(dt))
		
		logger.info("Minimizing into '{}'".format(destdir))
		memberfilepaths = self._readmembers() # Also fills self.committee!
		
		if os.path.exists(destdir):
			logger.warning("Destdir for minimization already exists, I will likely overwrite stuff!")
		else:
			os.makedirs(destdir)
		
		for (origfilepath, trainobj) in zip(memberfilepaths, self.committee):
			memberdestdirname = os.path.split(os.path.split(origfilepath)[0])[-1] # this is e.g. "member_010"
			memberdestdirpath = os.path.join(destdir, memberdestdirname)
			if not os.path.exists(memberdestdirpath):
				os.makedirs(memberdestdirpath)
			memberdestfilepath = os.path.join(memberdestdirpath, "Training.pkl")
			
			# And we make a tiny new training object, containing just the net
			net = trainobj.net
			net.resetcache()
			tinytrainobj = train.Training(net, None, name=trainobj.name)
			
			# To be able to select the best members later, we copy only the required stuff:
			tinytrainobj.optiterrs_val.append(trainobj.optiterrs_val[-1])
			tinytrainobj.optiterrs_train.append(trainobj.optiterrs_train[-1])
			tinytrainobj.optit = trainobj.optit
			
			# And save it
			tinytrainobj.save(memberdestfilepath)
			
		# We also copy the normers
		if os.path.exists(self.inputnormerpath):
			logger.info("Copying input normer...")
			shutil.copy(self.inputnormerpath, os.path.join(destdir, "input_normer.pkl"))
		if os.path.exists(self.targetnormerpath):
			logger.info("Copying target normer...")
			shutil.copy(self.targetnormerpath, os.path.join(destdir, "target_normer.pkl"))
		
		# And we write the config.
		configcopypath = os.path.join(destdir, self.name + ".cfg")
		with open(configcopypath, 'wb') as configfile:
			self.config.write(configfile)
		
		logger.info("Done with minimizing.")


	def predict(self, inputs):
		"""
		Checks the workdir and uses each network found or specified in config to predict some output.
		Does note care about the "Tenbilac" object used for the training!
		"""
		
		self._readmembers()
		
		
		if self.config.getboolean("predict", "selbest"):
			
			
			if self.config.get("predict", "selkind") == "bestn":
				thr = self.config.getint("predict", "thr")
				logger.info("Selecting {} best members among {}, using criterion '{}'.".format(thr, len(self.committee), self.config.get("predict", "bestkey")))
			elif self.config.get("predict", "selkind") == "sigmaclip":
				thr = self.config.getfloat("predict", "thr")
				logger.info("Sigma clipping to select the members. Selecting under {}sigma among {}.".format(thr, len(self.committee)))
			else:
				raise ValueError("Unknown member selection procedure")
			
			# We build a function to identify the best members
			bestkey = self.config.get("predict", "bestkey")
			if bestkey == "valerr":
				key = lambda trainobj: trainobj.optiterrs_val[-1]
			elif bestkey == "trainerr":
				key = lambda trainobj: trainobj.optiterrs_train[-1]
			elif bestkey == "nit":
				key = lambda trainobj: -1.0 * trainobj.optit # times minus one get the order right
			elif bestkey == "random":
				key = np.arange(len(self.committee))
				np.random.shuffle(key)
				for m, rid in zip(self.committee, key):
					m._rid = rid
				key = lambda trainobj: trainobj._rid
			else:
				raise ValueError("Unknown bestkey")
			logger.info("All potential committee members:")
			trainlistsummary(self.committee)
			
			if self.config.get("predict", "selkind") == "bestn":
				self.committee = sorted(self.committee, key=key)[0:thr]
			elif self.config.get("predict", "selkind") == "sigmaclip":
				keys = [key(mem) for mem in self.committee]
				_, ikeys = utils.sigma_clip_plus(keys, thr, get_indices=True)
				self.committee = [self.committee[i] for i in ikeys]
				logger.info("Sigma clipping to select the members. Selected {} members under {}sigma.".format(len(self.committee), thr))
			
		
			logger.info("Retained committee members:")
			trainlistsummary(self.committee)
				
		else:	
			logger.info("Keeping all members")
		
		
		# We norm the inputs
		if self.config.getboolean("norm", "oninputs"):
			logger.info("{}: normalizing inputs...".format(str(self)))
			input_normer = utils.readpickle(os.path.join(self.workdir, "input_normer.pkl"))
			inputs = input_normer(inputs)
			
		else:
			logger.info("{}: inputs do NOT get normed.".format(str(self)))
		
		# Run the predictions by all the committee members
		logger.info("{}: Making predictions with {} members...".format(str(self), len(self.committee)))
		predslist = [trainobj.net.predict(inputs) for trainobj in self.committee]
		
		# We might have to denorm the predictions:
		if self.config.getboolean("norm", "ontargets"):
			logger.info("{}: denormalizing predictions...".format(str(self)))
			target_normer = utils.readpickle(os.path.join(self.workdir, "target_normer.pkl"))
			predslist = [target_normer.denorm(preds) for preds in predslist]
		else:
			logger.info("{}: predictions do NOT get denormed.".format(str(self)))
		
		# And average the results
		logger.info("{}: Building predictions array...".format(str(self)))
		predsarray = np.ma.array(predslist)
		logger.info("Prediction shape is {}".format(predsarray.shape))
		assert predsarray.shape[0] == len(self.committee)

		combine = self.config.get("predict", "combine")
		if combine == "mean":
			retarray = np.mean(predsarray, axis=0)
		elif combine == "median":
			retarray = np.median(predsarray, axis=0)
		else:
			raise ValueError("Unknown combine")

		logger.info("{}: Done with averaging.".format(str(self)))
		return retarray
		


def trainlistsummary(trainlist):
	"""
	Logs a summary table describing a list of Training objects
	"""
	for trainobj in trainlist:
		nit = trainobj.optit
		#assert nit == len(trainobj.optiterrs_train)   # No, we don't need this, and it doesn't hold for minimized Trainings
		#assert nit == len(trainobj.optiterrs_val)
		trainerr = trainobj.optiterrs_train[-1]
		valerr = trainobj.optiterrs_val[-1]
		valerrratio = valerr / trainerr
		logger.info("{:>20}: {:5d} iterations, train = {:.6e}, val = {:.6e} ({:4.1f})".format(trainobj.name, nit, trainerr, valerr, valerrratio))



def datetimestr(dt):
	"""
	Returns a string that can be used in filenames
	"""
	nomicrodt = dt.replace(microsecond=0) # Makes format simpler
	return nomicrodt.isoformat().replace(":", "-")
	
	

def _trainworker(trainobj):
	"""
	Function that is mapped to a list of Training objects to perform the actual training on several CPUs.
	"""
	starttime = datetime.datetime.now()
	p = multiprocessing.current_process()
	logger.info("{} is starting to train with PID {} and kwargs {}".format(p.name, p.pid, trainobj._trainkwargdict))
	np.random.seed() # So that every worker uses different random numbers when selecting minibatches etc
	# Of course we DONT feed any static number to the seed, as we want it to be different every time!
	trainobj.opt(**trainobj._trainkwargdict)
	endtime = datetime.datetime.now()
	logger.info("{} is done, it took {}".format(p.name, str(endtime - starttime)))
	



	



"""
Helper functions for plots
Stuff that can be useful in many plots, to avoid repeading it...
"""

import numpy as np
import copy



def showselector(ax, sel):
	"""
	Writes text about the selector on the axes
	"""
	
	
	text = "\n".join([repr(crit) for crit in sel.criteria])
			
	ax.annotate(text, xy=(0.0, 0.5), xycoords='axes fraction', xytext=(8, -12), textcoords='offset points', ha='left', va='top')
	



def getrange(cat, feature):
	"""
	Returns a range (hi, low) to use when plotting the feature, respecting the feature's settings if not "None".
	So it makes sure you get actual numbers.
	"""


	if feature.low is not None and feature.high is not None: 
		customrange = (feature.low, feature.high)
	else:
		customrange = (np.min(cat[feature.colname]), np.max(cat[feature.colname]))

	return customrange




def summabin(x, y, xbinrange=(None, None), nbins=10, equalcount=False):
	"""
	For two 1D numpy arrays x and y, summarizes the y data in bins of x.
	
	Returns a dict with arrays containing different information for each bin:
	
	xbincents, xbinlows, xbinhighs,
	ymeans, ystds, ylowps, yhighps,
	ns
	
	
	"""
	
	# Some tests:
	assert x.ndim == 1
	assert y.ndim == 1
	assert x.size == y.size
	assert x.size > 1
	
	# Defining the bin range of the x axis:
	br = list(copy.deepcopy(xbinrange))
	if br[0] is None:
		br[0] = np.min(x)
	if br[1] is None:
		br[1] = np.max(x)
	
	# And now defining the bins:
	if equalcount is True:
		# With the latest numpy this should work without using a list comprehension, but ok...
		binlims = np.array([np.percentile(x, q) for q in np.linspace(0.0, 100.0, nbins+1)])
	else:
		binlims = np.linspace(br[0], br[1], nbins+1)
		
	bincenters = 0.5 * (binlims[0:-1] + binlims[1:])
	assert len(bincenters) == nbins
	
	binindices = np.digitize(x, bins=binlims)
	inrangeindices = np.arange(nbins)+1 # those indices that are within the binlims
	assert len(inrangeindices) == nbins
	

	# And now we loop over the bins to compute our stuff.
	
	ymeans = []
	ystds = []
	ylowps = []
	yhighps = []
	ns = []
		
	for ind in inrangeindices: # We loop over the bins
		
		inbools = binindices == ind # a boolean array
		nin = np.sum(inbools) # number of values falling into the current bin
		
		if nin < 2:
			ymeans.append(np.nan)
			ystds.append(np.nan)
			ylowps.append(np.nan)
			yhighps.append(np.nan)
			ns.append(np.nan)
			continue
		
		ns.append(nin)
		
		thesexvals = x[inbools]
		theseyvals = y[inbools]
		thisymean = np.mean(theseyvals)
				
		ymeans.append(thisymean)
		ystds.append(np.std(theseyvals))
		
		ylowps.append(np.fabs(np.percentile(theseyvals, 15.8) - thisymean))
		yhighps.append(np.fabs(np.percentile(theseyvals, 84.1) - thisymean))
		
	
	ret = {
		"xbincents":np.array(bincenters),
		"xbinlows":np.array(binlims[0:-1]),
		"xbinhighs":np.array(binlims[1:]),
		"ymeans":np.array(ymeans), 
		"ystds":np.array(ystds), 
		"ylowps":np.array(ylowps), 
		"yhighps":np.array(yhighps),
		"ns":np.array(ns)
	}	
	
	for key in ret.keys():
		assert len(ret[key]) == nbins
	
	return ret





"""
Attach the PSF measurements to the galaxy catalogs, and run the galaxy shape measurments.
Again individually for each subfields (but multi-cpu).
"""

import multiprocessing
import datetime
import numpy as np
import astropy.table

import momentsml
import momentsmlgreat3

import config
import measfcts

import logging
logging.basicConfig(format=config.loggerformat, level=logging.INFO)
logger = logging.getLogger(__name__)


incatfilepaths = []
outcatfilepaths = []


# Prepare the input catalogs:

for subfield in config.great3.subfields:

	config.great3.mkdirs(subfield)

	starcat = momentsml.tools.io.readpickle(config.great3.subpath(subfield, "obs", "star_meascat.pkl"))
	#print starcat
				
	incat = momentsmlgreat3.io.readgalcat(config.great3, subfield)
	
	
	# We add PSF info to this field. PSFs are already measured, and we take a random one (among 9) for each galaxy.
	
	starcat.meta = {} # Dump the "img" entry
	#matchedstarcat = starcat[np.zeros(len(incat), dtype=int)]
	matchedstarcat = starcat[np.random.randint(9, size=len(incat))]
	
	assert len(incat) == len(matchedstarcat)
	for colname in incat.colnames:
		if colname in matchedstarcat.colnames:
			raise RuntimeError("colname %s appears twice" % colname)
	
	incat = astropy.table.hstack([incat, matchedstarcat], join_type="exact", metadata_conflicts="error")
	
	#print incat.colnames
	#print incat[("ID", "x", "psfy", "psf_adamom_y", "psf_adamom_sigma")]
	
	# Add the reference to the img and psf stamps:
	
	incat.meta["img"] = momentsml.tools.imageinfo.ImageInfo(
		filepath=config.great3.galimgfilepath(subfield),
		xname="x",
		yname="y",
		stampsize=config.great3.stampsize(),
		workdir=config.great3.subpath(subfield, "obs", "img_measworkdir")
		)

	incat.meta["psf"] = momentsml.tools.imageinfo.ImageInfo(
		filepath=config.great3.starimgfilepath(subfield),
		xname="psfx",
		yname="psfy",
		stampsize=config.great3.stampsize(),
		workdir=None
		)

	# Write the input catalog
	incatfilepath = config.great3.subpath(subfield, "obs", "img_incat.pkl")
	momentsml.tools.io.writepickle(incat, incatfilepath)
	incatfilepaths.append(incatfilepath)
	
	# Prepare the filepath for the output catalog
	outcatfilepath = config.great3.subpath(subfield, "obs", "img_meascat.pkl")
	outcatfilepaths.append(outcatfilepath)


# And we're ready to run the measurements


# We pass some kwargs for the measfct
measfctkwargs = {"branch":config.great3}

# And we run with ncpu
momentsml.meas.run.general(incatfilepaths, outcatfilepaths, measfcts.gal, measfctkwargs=measfctkwargs,
					ncpu=config.great3.ncpu, skipdone=config.great3.skipdone)




"""
High-level functions to create a whole set of simulations.

These functions write their results to disk, and therefore **define a directory and
filename structure**, containing the different realizations, catalogs, etc.
This is very different from the lower-level functions such as those in stampgrid,
which do not specify any directory or filename structure.
"""

import os
import sys
import glob
import datetime
import tempfile
import cPickle as pickle
import numpy as np
import copy
import shutil
import multiprocessing
import astropy.table

import stampgrid
import inspect

from .. import tools

import logging
logger = logging.getLogger(__name__)


def multi(simdir, simparams, drawcatkwargs, drawimgkwargs=None,
	psfcat=None, psfselect="random", psfskipbad=False,
	ncat=2, nrea=2, ncpu=1, savetrugalimg=False, savepsfimg=False):
	"""
	Uses stampgrid.drawcat and stampgrid.drawimg to draw several (ncat) catalogs
	and several (nrea) "image realizations" per catalog.
	
	Supports multiprocessing (ncpu), and furthermore **generates additional simulations ** to any
	existing image set previously generated (instead of overwriting).
	To do so, unique filenames are generated (avoiding "race hazard") using ``tempfile``.
	So one could have **multiple processes running this function in parallel**, writing
	simulations using the same simparams in the same directory. Note that a timestamp is included
	in the filenames, but this is just to help human readability, not for uniqueness.
	
	About specifying the PSFs: this function takes a **single** psfcat, which has to link to an image
	via a psfcat.meta["img"] ImageInfo object. This psfcat does **not**
	need to have "n x n" rows (one for each galaxy in each catalog).
	Instead, for each of the n x n galaxy stamps in the ncat catalogs that will be created, a PSF from this psfcat
	will be assigned either randomly (psfchoice = "random") or in sequence (psfchoice = "sequential").
	In any case, the information about which PSF is used will be stored in the output catalogs (using the columns
	"psfx", "psfy", and meta["psf"]). In fact all columns of psfcat will be placed into the outputcat.
	Note that this function makes a copy of PSF image into its "simdir" (and refers to that copy instead of the
	image filepath given in psfcat), to be sure that all information to use the created simulations is kept inside
	simdir.
	
	:param simdir: Path to a directory where the simulations are written to.
		This directory does **not** have to be unique for every call of this function!
		Subdirectories reflecting the name of each simparams are made under the simdir.
		If the simparams directory already exists, simulations will be
		**added** to it, instead of overwriting any existing file.
	:param simparams: A sim.simparams instance that defines the distributions of parameters
	:param drawcatkwargs: Keyword arguments which will be directly passed to stampgrid.drawcat
	:type drawcatkwargs: dict
	:param drawimgkwargs: (currently not used) Idem, for stampgrid.drawimg.
		The filepaths simgalimgfilepath, simtrugalimgfilepath simpsfimgfilepath will be saved
		with auto-generated filenames, if asked for.
	:type drawimgkwargs: dict
	:param psfcat: catalog (or path to a pkl) describing the PSFs to use.
		It has to have meta["img"] set to link to the FITS image.
		If None, no PSF information will be passed to stampgrid.drawimg.
		If the psfcat is specified, data from **all its columns will be taken over** into
		the simulated catalogs (and not just the columns giving the x and y position of each PSF).
		However, any extra "meta" info from psfcat will not be taken over, as keeping it would be prone
		to confusions.
	:param psfselect: either "random" or "sequential", determines how the PSFs should be attributed
		to the simulated galaxies.
	:param psfskipbad: if True, skips PSFs from psfcat which have masked "psf_adamom_flux".
	:param ncat: The number of catalogs to be generated.
	:type ncat: int
	:param nrea: The number of realizations per catalog to be generated.
	:type nrea: int
	:param ncpu: Maximum number of processes to be used. Default is 1.
		Set to 0 for maximum number of available CPUs.
	:type ncpu: int
	:param savetrugalimg: if True, I will also save the true (unconvolved) galaxy images.
	:param savepsfimg: if True, I will also save the PSF stamps.
	
	
	As an illustration, an example of the directory structure that this function produces (for ncat=2, nrea=2)::
	
		% ls -1 simdir/name_of_simparams/*
		
		simdir/name_of_simparams/20141016T170441_kj34jj_psfimg.fits   <-- PSF image is copied!
		
		simdir/name_of_simparams/20141016T170441_1ZBNwd_cat.pkl  <-- the catalogs
		simdir/name_of_simparams/20141016T170441_BJjhps_cat.pkl  <-- (ncat different ones)

		simdir/name_of_simparams/20141016T170441_1ZBNwd_img:     <-- for each cat...
		20141016T170441_1ZBNwd_0_galimg.fits                     <-- ... the images 
		20141016T170441_1ZBNwd_0_trugalimg.fits
		20141016T170441_1ZBNwd_1_galimg.fits                     <-- (nrea different ones)
		20141016T170441_1ZBNwd_1_trugalimg.fits

		simdir/name_of_simparams/20141016T170441_BJjhps_img:
		20141016T170441_BJjhps_0_galimg.fits
		20141016T170441_BJjhps_0_trugalimg.fits
		20141016T170441_BJjhps_1_galimg.fits
		20141016T170441_BJjhps_1_trugalimg.fits

	
	For a single call to this function, the timestamp is the same for all catalogs. This is
	handy if you want to delete all files from a particular call.
	
	Note that the unique filename of a catalog is repeated in the filename of every single
	realization image.  This is intended, so that you can collect things that are based on
	realization image filenames made from a single name_of_simparams into one directory.
	It also makes things safer.
	
	.. note::
		About the ImageInfo of the returned catalogs: as each catalog has nrea realizations, we cannot simply attach
		a single meta["img"] ImageInfo object to a catalog. We have to define a system for storing "realizations".
		For now, what we do is writing a list of ImageInfo objects into meta["imgreas"]. We can talk about this...
	
	
	"""
	starttime = datetime.datetime.now()
	prefix = starttime.strftime("%Y%m%dT%H%M%S_")
	if drawimgkwargs is None:
		drawimgkwargs = {}
	
	# First, some general checks:
	if ncat < 1 or nrea < 1:
		raise RuntimeError("ncat and nrea must be above 0")
	
	# Some tests about the drawimgkwargs:
	forbiddendrawimgkwargs = ["simgalimgfilepath", "simtrugalimgfilepath", "simpsfimgfilepath"]
	for kwarg in forbiddendrawimgkwargs:
		if kwarg in drawimgkwargs:
			raise RuntimeError("I can't respect your drawimgkwarg '%s', do not specify it." % (kwarg))
			# Better *not* modify the drawimgkwargs as was done below (side effect !)
			#logger.warning("I will not respect your drawimgkwarg '%s'" % (kwarg))
			#drawimgkwargs.pop(kwarg)

	# Preparing a workdir
	workdir = os.path.join(simdir, simparams.name)
	if not os.path.exists(workdir):
		os.makedirs(workdir)
		logger.info("Creating a new set of simulations named '%s'" % (simparams.name))
	else:
		logger.info("Adding new simulations to the existing set '%s'" % (simparams.name))

	logger.info("Drawing %i catalogs, and %i image realizations per catalog" % (ncat, nrea))
	logger.info("All files will be written into '%s', with prefix '%s'" % (workdir, prefix))
	
	# We create the catalogs, this is fast and there is no need to parallelize this: 
	logger.info("Drawing galaxy catalogs...")
	catalogs = [stampgrid.drawcat(simparams, **drawcatkwargs) for i in range(ncat)]
	
		
	# We now attribute PSFs to each source in these catalogs. That's a little bit of work.
	# We have to be careful not to mix PSFs within blocks of shape noise cancellation.
	if psfcat is not None:
	
		if (psfskipbad is True) and ("psf_adamom_flux" in psfcat.colnames):
			logger.info("We kick out PSFs that could not be measured!")
			psfcat = psfcat[psfcat["psf_adamom_flux"].mask == False]
			logger.info("{} PSFs remain".format(len(psfcat)))
	
		# First we check that the PSF stuff looks fine:
		if "img" not in psfcat.meta:
			raise RuntimeError("ImageInfo of psfcat is missing, put it into psfcat.meta['img']")
		psfcat.meta["img"].checkcolumns(psfcat)
	
		# Now we copy the PSF image into the workdir.
		# We want this to be perfectly collision-free -- just using the timestamp prefix is *not* enough,
		# as several instances of this multi() could be launched at the same time.
		# So we use tempfile to build a filename for this PSF image.
		
		drawpsfimgfile = tempfile.NamedTemporaryFile(mode='wb', prefix=prefix, suffix="_psfimg.fits", dir=workdir, delete=False)
		drawpsfimgfilename = drawpsfimgfile.name
		drawpsfimgfile.close()	
		drawpsfimgfilepath = os.path.join(workdir, drawpsfimgfilename)
		logger.info("Copying PSF image into the simdir, with filename '%s'..." % (drawpsfimgfilename))
		shutil.copy(psfcat.meta["img"].filepath, drawpsfimgfilepath)
	
		# We can now prepare the catalog.meta["psf"] ImageInfo for each catalog, to point to that FITS file.
		# It's almost the same as psfcat.meta["img"], except that the filepath should point to our new copy.
		# Let's make a new object instead of tweaking the existing one, so that "name" gets correctly set etc.
		
		psfinfo = tools.imageinfo.ImageInfo(
			drawpsfimgfilepath,
			xname=psfcat.meta["img"].xname,
			yname=psfcat.meta["img"].yname,
			stampsize=psfcat.meta["img"].stampsize,
			pixelscale=getattr(psfcat.meta["img"], "pixelscale", 1.0) # using getattr in case you apply it to old cats
			)
			
		# Now, we have to populate the catalog itself with PSF coordinates.
		# This part is to be done independently for each catalog, so we loop over the catalogs:
		
		logger.info("Attributing PSFs using psfselect='%s'" % (psfselect))
		mergedcatalogs = []
		for catalog in catalogs:
		
			# We set the catalog.meta["psf"] ImageInfo.
			catalog.meta["psf"] = psfinfo # It's the same for all catalogs.
			if psfselect == "random":
				# We repeat the random drawing for each catalog, as we want different PSFs for each catalog.
				if catalog.meta["snc_type"] == 0:		
					matched_psfcat = psfcat[np.random.randint(low=0, high=len(psfcat), size=len(catalog))] # the "high" of randint is exclusive !
				
				else:
					assert len(catalog) % catalog.meta["nsnc"] == 0
					nsncblocks = len(catalog) / catalog.meta["nsnc"] # The number of different PSFs that we need.
					randompsfcatindices = np.repeat(np.random.randint(low=0, high=len(psfcat), size=nsncblocks), catalog.meta["nsnc"])
					assert randompsfcatindices.ndim == 1
					assert len(randompsfcatindices) == len(catalog)
					
					matched_psfcat = psfcat[randompsfcatindices]
					assert len(matched_psfcat) == len(catalog)	
	
			elif psfselect == "sequential":
				raise RuntimeError("Not yet implemented")
				
			else:
				raise RuntimeError("Unknown psfselect")	
			
			# Now that we have a copy of psfcat (to be merged with catalog),
			# we can dump its metadata without any side-effects:
			matched_psfcat.meta = {}
			
			# And we merge this matched_psfcat into the catalog.
			# However we want to be sure that columns do not have conflicting names.
			# So we first test for this:
			for colname in psfcat.colnames:
				if colname in catalog.colnames:
					raise RuntimeError("The column '%s' of psfcat is in conflict with the drawcat columns")
			
			# The merge itself:
			mergedcatalog = astropy.table.hstack([catalog, matched_psfcat], join_type='exact',
				table_names=['gal', 'psf'], uniq_col_name='{table_name}_{col_name}', metadata_conflicts="error")
			
			mergedcatalogs.append(mergedcatalog)
		
		catalogs = mergedcatalogs # We don't need the non-merged catalogs anymore.

	# Now we prepare the catalogs one by one by adding the ImageInfos of the nrea realizations to be drawn
	# for each catalog.
	# For this, we need filenames, and we want them to be unique.
	# This is not done with the timestamp!  The timestamp is only here to help humans.
	# The module tempfile takes care of making the filename unique.
	# And so, we do this in one loop
	
	for catalog in catalogs:
	
		# We open a file object:
		catfile = tempfile.NamedTemporaryFile(mode='wb', prefix=prefix, suffix="_cat.pkl", dir=workdir, delete=False)
		
		# Now we can get the unique filename
		catalog.meta["catname"] = os.path.basename(str(catfile.name)).replace("_cat.pkl","")
		
		# The images will be written here:
		catimgdirpath = os.path.join(workdir, catalog.meta["catname"] + "_img")
		
		# We set the ImageInfos of the reas, using this unique catname
		catalog.meta["imgreas"] = [
			tools.imageinfo.ImageInfo(
				os.path.join(catimgdirpath, "%s_%i_galimg.fits" % (catalog.meta["catname"], reaindex)),
				xname="x", yname="y", stampsize = catalog.meta["stampsize"], pixelscale=catalog.meta['pixelscale'])
			for reaindex in range(nrea)]
		
		# And we can write this catalog to disk
		pickle.dump(catalog, catfile) # We directly use this open file object.
		catfile.close()
		logger.info("Wrote catalog '%s'" % catalog.meta["catname"])
		
	
	# And now we draw the image realizations for those catalogs.
	# This is done with multiprocessing.
	# We make a multiprocessing pool map over all combinations of catalogs and realizations,
	# as we want this to be efficient for both (1 cat, 20 rea), and (20 cat, 1 rea)
	# For this, we prepare a flat list of _WorkerSettings objects for all (cat, rea) combinations,
	# and run a pool of _worker functions on this list.
	
	wslist = []
	for catalog in catalogs:	
		for reaindex in range(nrea):
			
			# We have to customize the drawimgkwargs, and so we work on a copy
			thisdrawimgkwargs = copy.deepcopy(drawimgkwargs)
				
			# Preparing the filepaths in which we will write the output image(s)
			catname = catalog.meta["catname"]
			catimgdirpath = os.path.join(workdir, catname + "_img")
				
			# We have already stored the simgalimgfilepath in the catalog meta:
			thisdrawimgkwargs["simgalimgfilepath"] = catalog.meta["imgreas"][reaindex].filepath	
			#thisdrawimgkwargs["simgalimgfilepath"] =\
			#	os.path.join(catimgdirpath, "%s_%i_galimg.fits" % (catname, reaindex))
		
			# If the user asked for a trugalimg and a psfimg, we also prepare these filepaths.
			if savetrugalimg:
				thisdrawimgkwargs["simtrugalimgfilepath"] = os.path.join(catimgdirpath, "%s_%i_trugalimg.fits" % (catname, reaindex))
			if savepsfimg:
				thisdrawimgkwargs["simpsfimgfilepath"] = os.path.join(catimgdirpath, "%s_%i_psfimg.fits" % (catname, reaindex))
	
			ws = _WorkerSettings(catalog, reaindex, thisdrawimgkwargs, workdir)
			
			wslist.append(ws)
		
		# While in this simple loop over catalogs, we also make the dir that will contain the images.
		# Indeed the worker loop iterates over this same dir for every realization.
		
		os.mkdir(os.path.join(workdir, catalog.meta["catname"] + "_img"))
		
		
	assert len(wslist) == ncat * nrea
	
	# The catalogs could be heavy, but note that we do not put unique copies of the catalogs in this list !
	# Still, it would seem better to just have small thinks like "indexes" in the settings.
	# However it seems that accessing shared memory from a multiprocessing.Pool is not trivial.
	# So until we need something better, we leave it like this.
	# Note for the future: instead of thinking about how to share memory to optimize this, the workers could well
	# read their data from disk, and stay embarassingly parallel. 

	if ncpu == 0:
		try:
			ncpu = multiprocessing.cpu_count()
		except:
			logger.warning("multiprocessing.cpu_count() is not implemented!")
			ncpu = 1
			
	logger.info("Start drawing %i images using %i CPUs" % (len(wslist), ncpu))

	# Suppress the info-or-lower-level logging from the low-level functions:
	#stampgridlogger = logging.getLogger("momentsml.sim.stampgrid")
	#stampgridlogger.setLevel(logging.WARNING)
	
	if ncpu == 1:
		# The single-processing version (not using multiprocessing to keep it easier to debug):
		logger.debug("Not using multiprocessing")
		map(_worker, wslist)

	else:
		# multiprocessing map:
		pool = multiprocessing.Pool(processes=ncpu)
		pool.map(_worker, wslist)
		pool.close()
		pool.join()
	
	endtime = datetime.datetime.now()
	nstamps = len(catalogs[0])*nrea
	logger.info("Done, the total time for drawing %i stamps was %s" % (nstamps, str(endtime - starttime)))
	logger.info("That's %.3f ms per stamp (with ncpu = %i)." % (1e3*(endtime - starttime).total_seconds()/nstamps, ncpu))
	
	
	# The current logfile (below) is not that useful. Rethink it if needed.
	"""
	# The catalogs are drawn, we save a log file about this
	logger.debug("Now writing logfile...")
	logfilepath = os.path.join(workdir, prefix + "log.txt")
	logfile = open(logfilepath, "w")
	logfile.write("Logfile of momentsml.sim.run.multi, written to %s\n\n" % (logfilepath))
	logfile.write(str(simparams) + "\n")
	logfile.write("drawcatkwargs: " + repr(drawcatkwargs) + "\n")
	logfile.write("drawimgkwargs: " + repr(drawimgkwargs) + "\n")
	# I did not manage to save the code of simparams, here is my attempt:
	#logfile.write("\n")
	#logfile.write(str(simparams) + ":\n")
	#logfile.write(inspect.getsourcelines(type(simparams)))
	logfile.close()
	logger.debug("Done with writing logfile")
	"""


class _WorkerSettings():
	"""
	A class to hold together all the settings for processing a catalog-realization combination.
	If one day we have different drawimg() functions, we'll just pass this function here as well.
	"""
	
	def __init__(self, catalog, reaindex, drawimgkwargs, workdir):
		"""
		The catalog's catname, reaindex, and workdir define the filepaths in which the image(s)
		drawn with the drawimgkwargs will be written.
		"""
		
		self.catalog = catalog # No copy needed, we won't change it!
		self.reaindex = reaindex
		self.drawimgkwargs = drawimgkwargs # This is already a changed deep copy from the original argument to multi().
		self.workdir = workdir # Stays the same for all workers !
	
		
	def __str__(self):
		"""
		A short string describing these settings
		"""
		return "[catalog '%s', realization %i]" % (self.catalog.meta["catname"], self.reaindex)
	
	
def _worker(ws):
	"""
	Worker function that processes one _WorkerSettings object.
	"""		
	starttime = datetime.datetime.now()
	# Make sure that we indeed start from different seeds
	np.random.seed()
	p = multiprocessing.current_process()
	logger.info("%s is starting to draw %s with PID %s" % (p.name, str(ws), p.pid))
	
	# It's just a single call:
	stampgrid.drawimg(ws.catalog, **ws.drawimgkwargs)
	
	endtime = datetime.datetime.now()
	logger.info("%s is done, it took %s" % (p.name, str(endtime - starttime)))



"""
Functions to help visualizing postage stamps of selected sources in large images.
"""

from .. import tools 
import logging
logger = logging.getLogger(__name__)

import numpy as np

try:
	import f2n
except ImportError:
	logger.warning("Could not import f2n, no problem if you don't use it. You can find it here: http://obswww.unige.ch/~tewes/f2n_dot_py/")


def pngstampgrid(img, cat, pngfilepath, xname="x", yname="y", stampsize=100, ncols=5, upsample=4, z1="auto", z2="auto"):
	"""
	For this it uses the (slightly outdated) f2n.py module.

	:param img: either a galsim image or the filepath to a FITS image
	:param cat: an astropy table
	:param pngfilepath: the png file path to be written
	:param xname: colname for the x position (in pixel)
	:param yname: colname for y
	:param stampsize: stamp size (in pixels) to be extracted
	:param ncols: number of postage-stamp columns
	:param upsample: postage-stamp upsample rate
	:param z1: "z" scale low
	:param z2: "z" scale high
	
	"""
	
	
	if type(img) is str:
		logger.debug("Filepath given, loading the image...")
		img = tools.image.loadimg(img)
		
	n = len(cat)
	nrows = int(np.ceil(float(n)/float(ncols)))
	logger.info("Preparing %i x %i stamps of %i x %i pixels each..." % (ncols, nrows, stampsize, stampsize))	
	
	stamprows = []
	for nrow in range(nrows):
		stamprow = []
		for ncol in range(ncols):
			
			index = ncol + ncols*nrow
			if index < n: # Then we have a galaxy to show
				gal = cat[index]
				(x, y) = (gal[xname], gal[yname])
				(gps, flag) = tools.image.getstamp(x, y, img, stampsize)
				npstamp = gps.array
				
				f2nstamp = f2n.f2nimage(numpyarray=npstamp, verbose=False)

				f2nstamp.setzscale(z1, z2)
				f2nstamp.makepilimage("log", negative = False)
				f2nstamp.upsample(upsample)
			
				txt = [
					"%i (%i, %i)" % (gal.index, x, y),
				]
				f2nstamp.writeinfo(txt, colour=255)
				
				# Crosshair:
				s = stampsize
				f2nstamp.drawline(s/2, s/4, l=30, t=np.pi/2.0)	
				f2nstamp.drawline(s/4, s/2, l=30, t=0.0)

				# Just for reference, some other stuff from previous MomentsML versions:				
				#f2nstamp.drawrectangle(1, s-1, 1, s-1)	
				#f2nstamp.drawrectangle(140, s-140, 140, s-140, colour=(0,255,255))	
				# Showing the measured shape, in red
				#e = np.hypot(galaxy.gal_e1, galaxy.gal_e2)
				#t = 0.5*np.arctan2(galaxy.gal_e2, galaxy.gal_e1)
				#f2nstamp.drawline(x = s/4, y=s/4 , l=150*e, t=t, width=3, colour=(255,0,0))


			else: # No more galaxies, we just fill the splot with a grey empty stamp.
				npstamp = np.zeros((stampsize, stampsize))
				f2nstamp = f2n.f2nimage(numpyarray=npstamp, verbose=False)
				f2nstamp.setzscale(-1.0, 1.0)
				f2nstamp.makepilimage("lin", negative = False)
				f2nstamp.upsample(4)
			
			
			stamprow.append(f2nstamp)
		stamprows.append(stamprow)
	f2n.compose(stamprows, pngfilepath)
	logger.info("Wrote %s" % (pngfilepath))
	
	
#def npstampgrid(img, catalog, xname="x", yname="y", stampsize=100):
#	"""
#	I build a numpy array with stamps, intended for visualization
#	"""
#
#	#n = len(catalog)
#	#nrows = int(np.ceil(n/10))
#	#big = np.zeros((10*stampsize, nrows*stampsize))
#	#print n, nrows
#	
#	stamplist = []
#	for gal in catalog:
#		(x, y) = (gal[xname], gal[yname])
#		(gps, flag) = getstamp(x, y, img, stampsize)
#		if flag != 0:
#			stamplist.append(np.zeros(stampsize, stampsize))
#		else:
#			stamplist.append(gps.array)
#	
#	big = np.vstack(stamplist)
#	return big





import numpy as np
import tenbilac
import pylab as plt

import logging
logging.basicConfig(level=logging.INFO)

#np.random.seed(10)

net1 = tenbilac.net.Net(ni=1, nhs=[3])
net2 = tenbilac.net.Net(ni=1, nhs=[4])
net3 = tenbilac.net.Net(ni=1, nhs=[2,2])

members = [net1, net2, net3]

comm = tenbilac.committee.Committee(members)

n = 200 # Size of training set
noise_scale = 0.05
ntest = 100 # Number of points to draw the regression line

def f(x):
	return np.sin(x)/(2. + x)

params = np.linspace(0.0, 10.0, 2*n).reshape((1, 2*n))
np.random.shuffle(params.flat) # Such that we can seperate the training set afterwards
obs = f(params) + noise_scale*np.random.randn(2*n).reshape((1, 2*n))

# Tenbilac only accept 3D input arrays
params = np.array([params])

# Normer object for the inputs and targets
obs_normer = tenbilac.data.Normer(obs)
params_normer = tenbilac.data.Normer(params)

# Do the normalisation
normobs = obs_normer(obs)
normparams = params_normer(params)

# Create some points where we want to see the value of the fitted obs, given the data.
testparams = np.linspace(-1.0, 11, ntest).reshape((1, ntest))
normtestparams = params_normer(testparams)

# Prepare different traindata objects
dat1 = tenbilac.data.Traindata(inputs=normparams[:,:,:n], targets=normobs[:,:n], valfrac=0.2)
dat2 = tenbilac.data.Traindata(inputs=normparams[:,:,n:], targets=normobs[:,n:], valfrac=0.3)
dat = [dat2, dat1, dat2]

# Preparing the training: 
# Tell tenbilac that you want to train with different trainings
ctraining = tenbilac.committee.CommTraining(comm, dat=dat, errfctname=["mse", "mse", "mse"], multiple_trainings=True)

# The method `call` allows to call a function on each of the member of the class
# In the simplereg.py example this would be training.net.addnoise()
# ctraining.call(attr='net', method='addnoise', wscale=0.3, bscale=0.3)
# you can also do something like this to customise the training:
wscales = [0.25, 0.3, 0.35]
[t.net.addnoise(wscale=wscales[i], bscale=0.3) for i, t in enumerate(ctraining.trainings)]

# In the simplereg.py example this would be training.bfgs()
ctraining.call(method='opt', call_ncpu=3, maxiter=500, algo="bfgs")

# Predicting the testparams, notice the exact same call method.
normtestpreds = comm.call(method='run', inputs=normtestparams)
testpreds = obs_normer.denorm(normtestpreds)

# We go back from a 3D array to 2D for plottting
params = params[0]

fig = plt.figure()

ax = fig.add_subplot(1, 1, 1)
ax.plot(params.T, obs.T, "b.", label="obs")
for ii in range(len(comm.members)): 
	ax.plot(testparams.T, testpreds[ii].T, label='%i: %s' % (ii, comm.members[ii]))

ax.plot(testparams.T, np.mean(testpreds.T, axis=2), lw=2, label='mean fit')
ax.plot(testparams.T, f(testparams.T), "k--", lw=2, label="truth")
ax.set_xlabel(r"$x$", fontsize=18)
ax.set_ylabel(r"$y = f(x) + \mathrm{noise}$", fontsize=18)
ax.set_ylim([-0.5, 0.5])
ax.legend(loc='best')

plt.tight_layout()

plt.show()	

print 'test completed.'



import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os

import plot_3_val_s as plot

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)

	
# Getting the path to the correct directories
traindir = os.path.join(config.traindir, config.datasets["tp"])
	
# And to the validation catalogue
valcatpath = os.path.join(config.simmeasdir, config.datasets["vp"], "groupmeascat.pkl")
predcatpath = os.path.join(config.valdir, config.valname + ".pkl")


# The prediction:

cat = momentsml.tools.io.readpickle(valcatpath)

"""
select = False # prefer selecting at the plot stage, not here.

if select:
	#momentsml.tools.table.addstats(cat, "snr")

	s = momentsml.tools.table.Selector("fortrain", [
		#("max", "adamom_failfrac", 0.01),
		#("min", "tru_rad", 3.0),
		#("min", "snr_mean", 10.0),
	])
	cat = s.select(cat)
"""

predcat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist, traindir)

momentsml.tools.io.writepickle(predcat, predcatpath)


# The quick automatic plot:
"""	
for conf in config.shearconflist: # For the plot, we go through them one by one here.
	
	# We read the file, to get a fresh one for each component:
	predcat = momentsml.tools.io.readpickle(predcatpath)

	
	# And make the plot	
	if "s2" in conf[0] or "g2" in conf[0]:
		component = 2
	else:
		component = 1
	
	valfigpath = os.path.join(config.valdir, valname + "_component{}".format(component) + ".png")

	plot.plot(predcat, component, mode="s", filepath=valfigpath)
"""



"""
Measures features of the PSFs, 
individually for each subfield.
"""
import multiprocessing
import datetime
import numpy as np
import astropy.table

import momentsml

import config
import measfcts

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


for subfield in config.great3.subfields:

	config.great3.mkdirs(subfield)

	# We don't bother reading the starcat, and just make one
	stars = []
	for i in range(3):
		for j in range(3):
			stars.append( [0.5 + config.great3.stampsize()/2.0 + i*config.great3.stampsize(), 0.5 + config.great3.stampsize()/2.0 + j*config.great3.stampsize()] )
	stars = np.array(stars)
	
	starcat = astropy.table.Table([stars[:,0], stars[:,1]], names=('psfx', 'psfy'))
	#print starcat
	
	# We add the subfield number to the catalog, might be handy later.
	starcat["subfield"] = subfield
	
	
	# To measure the stars, we attach the image:
	starcat.meta["img"] = momentsml.tools.imageinfo.ImageInfo(
		filepath=config.great3.starimgfilepath(subfield),
		xname="psfx",
		yname="psfy",
		stampsize=config.great3.stampsize(),
		workdir=config.great3.subpath(subfield, "obs", "star_measworkdir")
		)

	starcat = measfcts.psf(starcat, branch=config.great3)
	#print starcat[["psfx", "psfy", "psf_sewpy_XWIN_IMAGE", "psf_sewpy_YWIN_IMAGE", "psf_adamom_x", "psf_adamom_y"]]
	#print starcat

	momentsml.tools.io.writepickle(starcat, config.great3.subpath(subfield, "obs", "star_meascat.pkl"))
	




"""
A layer holds the parameters (weights and biases) to compute the ouput of each of its neurons based on the input of the previous layer.
This means that the "input" of a neural network is *not* such a Layer object.
The "first" real Layer in a network is in fact the first hidden layer.

"""


import numpy as np

import logging
logger = logging.getLogger(__name__)

from . import act


		
class Layer():	
	def __init__(self, ni, nn, actfct=act.tanh, name="None", mode="sum"):
		
		"""
		:param ni: Number of inputs
		:param nn: Number of neurons
		:param actfct: Activation function
		:param mode: "sum" or "mult". If sum, the neurons perform the usual linear combinations.
			If mult, they work as "Product Units" (Durbin Rumelhart 1989).
		"""

		self.ni = abs(ni)
		self.nn = abs(nn) # We keep those positive, even if negative might have indicated "mult" earlier in the code.
		self.actfct = actfct
		self.name = name
		
		self.mode = mode
		if not self.mode in ("sum", "mult"):
			raise RuntimeError("Unknown layer mode")

		self.weights = np.zeros((self.nn, self.ni)) # first index is neuron, second is input
		self.biases = np.zeros(self.nn) # each neuron has its bias
		
		# We will store the last "run" of the layer in a cache:
		self.resetcache()
	
	def resetcache(self):
		"""
		Purges the cache, to save memory (typically before saving...)
		"""
		self.cache_lastinputs = None
		self.cache_lastweights = None
		self.cache_lastbiases = None
		self.cache_lastoutputs = None
	
		
	def __str__(self):
		return "Layer '{self.name}', mode {self.mode}, ni {self.ni}, nn {self.nn}, actfct {self.actfct.__name__}".format(self=self)
	
	def modecode(self):
		"""
		Just a little "*" to label unconventional layers in logs etc
		"""
		if self.mode == "sum":
			return ""
		else:
			return "*"
	
	def addnoise(self, wscale=0.1, bscale=0.1, multwscale=0.1, multbscale=0.1):
		"""
		Adds some noise to weights and biases
		"""
		if self.mode == "sum":
			self.weights += wscale * np.random.randn(self.weights.size).reshape(self.weights.shape)
			self.biases += bscale * np.random.randn(self.biases.size)
		elif self.mode == "mult":
			self.weights += multwscale * np.random.randn(self.weights.size).reshape(self.weights.shape)
			self.biases += multbscale * np.random.randn(self.biases.size)
		else:
			raise RuntimeError("Unknown mode")
		
	
	def setzero(self):
		"""
		Sets all weights and biases to zero
		"""
		logger.info("Setting {self.mode}-layer '{self.name}' parameters to zero...".format(self=self))
		#self.weights = np.zeros(self.weights.shape) # No, this does not work! It kills the get_params reference used to set up the Training *before* noise is added.
		#self.biases = np.zeros(self.biases.shape) # Leave this commented to prevent later attempts to change it.
		self.weights *= 0.0
		self.biases *= 0.0
	

	def setidentity(self, onlyn=None):
		"""
		Sets the weights and biases so that this layer transports all its first nn inputs to the nn outputs.
		This is meant to give a promising initial condition for training a calibration-like task. 
		
		:param onlyn: If set and positive, limits the number of neurons to be set. Otherwise, all are set.
		:type onlyn: int
		"""
		self.setzero()
		
		npossible = min(self.ni, self.nn) # Can't transport more inputs
		if onlyn is None or onlyn < 0:
			ngofor = npossible
		else:
			if onlyn > npossible:
				logger.warning("Cannot setidentity for more neurons (onlyn={}) than possible ({})".format(onlyn, npossible))
			ngofor = min(onlyn, npossible)
		if ngofor < 0:
			raise RuntimeError("Cannot set identity, something wrong with number of neurons.")
		
		for n in range(ngofor):
			self.weights[n, n] = 1.0
		logger.info("Setting identity for {}: it now transports {} inputs".format(str(self), ngofor))
		
	
	
	def report(self):
		"""
		Returns a text about the weights and biases, useful for debugging
		"""
		
		txt = []
		txt.append(str(self) + ":")
		for inn in range(self.nn):
			if self.mode == "sum":
				txt.append("    output {inn} = {act} ( input * {weights} + {bias} )".format(
					inn=inn, act=self.actfct.__name__, weights=self.weights[inn,:], bias=self.biases[inn]))
			elif self.mode == "mult":
				txt.append("    output {inn} = {act} ( sign * prod (input ** {weights}) +(???) {bias} )".format(
					inn=inn, act=self.actfct.__name__, weights=self.weights[inn,:], bias=self.biases[inn]))
		return "\n".join(txt)
	
	
	def nparams(self):
		"""
		Retruns the number of parameters of this layer
		"""
		return self.nn * (self.ni + 1)
	
		
	def run(self, inputs):
		"""
		Computes output from input, as "numpyly" as possible, using only np.dot (note that np.ma.dot does not
		work with 3D masked arrays, and np.tensordot seems not available for masked arrays.
		This means that np.dot determines the order of indices, as following.
		
		If inputs is 1D,
			the input is just an array of features for a single case
			the output is a 1D array with the output of each neuron
			
		If inputs is 2D,
			input indices: (feature, case)
			output indices: (neuron, case) -> so this can be fed into the next layer...
		
		If inputs is 3D, 
			input indices: (realization, feature, case)
			output indices: (realization, neuron, case)
			
		"""
		
		
		# Cache: if nothing has changed, no need to run... (no matter the "mode"):
		
		if self.cache_lastoutputs is None:
			# We have no cache (first iteration), so nothing to do but run...
			pass
		else:
			# We check if anything has changed, starting with the fast things:
			if np.all(self.weights == self.cache_lastweights):
				if np.all(self.biases == self.cache_lastbiases):
					if np.all(inputs == self.cache_lastinputs):
						# I was thinking of using np.all_close() instead of "==" here...
						# But in fact, with the previous layers being cached as well,
						# using "==" seems just the right thing to do, nice!
					
						#logger.info("BINGO for mode {} !!!".format(self.mode))
						
						# Nothing has changed, so we can return the cached outputs:
						return self.cache_lastoutputs
		
		#logger.info("No BINGO for mode {}".format(self.mode))
		
		if self.mode == "sum":
			if inputs.ndim == 1:
				return self.actfct(np.dot(self.weights, inputs) + self.biases)
		
			elif inputs.ndim == 2:
				assert inputs.shape[0] == self.ni		
				return self.actfct(np.dot(self.weights, inputs) + self.biases.reshape((self.nn, 1)))
		
			elif inputs.ndim == 3:
				assert inputs.shape[1] == self.ni
			
				# Doing this:
				# self.actfct(np.dot(self.weights, inputs) + self.biases.reshape((self.nn, 1, 1)))
				# ... gives ouput indices (neuron, realization, case)
				# We need to change the order of those indices:
				
				self.cache_lastoutputs = np.rollaxis(self.actfct(np.dot(self.weights, inputs) + self.biases.reshape((self.nn, 1, 1))), 1)
				self.cache_lastinputs = inputs.copy()
				self.cache_lastweights = self.weights.copy()
				self.cache_lastbiases = self.biases.copy()
				return self.cache_lastoutputs
			
				# Note that np.ma.dot does not work for 3D arrays!
				# We do not care about masks at all here, just compute assuming nothing is masked.
			
			else:
				raise RuntimeError("Input shape error")

# The initial implementation of mult, works for positive inputs:
#		elif self.mode == "mult":
#			
#			if inputs.ndim == 1:
#				return self.actfct(np.prod(np.power(inputs, self.weights), axis=1) + self.biases) # Phew, that was easy, nice!
#		
#			elif inputs.ndim == 2:
#				assert inputs.shape[0] == self.ni			
#				return self.actfct(np.transpose(np.prod([np.power(inputs[:,i], self.weights) for i in range(inputs.shape[1])], axis=2) + self.biases))
#				
#			elif inputs.ndim == 3:
#				assert inputs.shape[1] == self.ni
#				return np.swapaxes(self.actfct(np.prod([[np.power(inputs[j,:,i], self.weights) for i in range(inputs.shape[2])] for j in range(inputs.shape[0])], axis=3) + self.biases), 1, 2)
#				
#			else:
#				raise RuntimeError("Input shape error")
			
# new variants for negative and positive inputs
		elif self.mode == "mult":
			
			signlim = 0.5
			self.biases *= 0.0 # For now I kill those: biases have no effect. Remains to be explored how to use them best.
			
			if inputs.ndim == 1:		
				# The test used as exponent has to be False if we want to ignore the sign, True to use it
				signs = np.prod(np.power(np.sign(inputs), np.fabs(self.weights) > signlim), axis=1)
				return signs * self.actfct(np.prod(np.power(np.fabs(inputs), self.weights), axis=1) + self.biases) # Phew, that was easy, nice!
		
			elif inputs.ndim == 2:
				assert inputs.shape[0] == self.ni
				signs = np.transpose(np.prod([np.power(np.sign(inputs[:,i]), np.fabs(self.weights) > signlim) for i in range(inputs.shape[1])], axis=2))	
				return signs * self.actfct(np.transpose(np.prod([np.power(np.fabs(inputs[:,i]), self.weights) for i in range(inputs.shape[1])], axis=2) + self.biases))
				
			elif inputs.ndim == 3:
				assert inputs.shape[1] == self.ni
				signs = np.swapaxes(np.prod([[np.power(np.sign(inputs[j,:,i]), np.fabs(self.weights) > signlim) for i in range(inputs.shape[2])] for j in range(inputs.shape[0])], axis=3), 1, 2)
				self.cache_lastoutputs = signs * np.swapaxes(self.actfct(np.prod([[np.power(np.fabs(inputs[j,:,i]), self.weights) for i in range(inputs.shape[2])] for j in range(inputs.shape[0])], axis=3) + self.biases), 1, 2)
				self.cache_lastinputs = inputs.copy()
				self.cache_lastweights = self.weights.copy()
				self.cache_lastbiases = self.biases.copy()
				return self.cache_lastoutputs
				
			else:
				raise RuntimeError("Input shape error")
		
		
		
		
		
	




import config
import measfcts
import os
import numpy as np
import matplotlib.ticker as ticker

import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)


#momentsml.plot.figures.set_fancy(14)
from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)


###############################

select = True # will be set to False if useweights is True

useweights = False # <--- to be used with a "VP" like structure in place of "VO".
regressmethod = 1 # Only matters if useweights.
# For a VP-like dataset, it only makes sense to use regressmethod1 here, as we want the weights to act over the different true galaxies within each bin.
# Regressmethod 2 would kill the effect of weights, only use it if cases contain different gals.

###############################

if useweights is False:
	valname = config.valname
else:
	valname = config.wvalname
	select = False

valcat = os.path.join(config.valdir, valname + ".pkl")
cat = momentsml.tools.io.readpickle(valcat)
momentsml.tools.table.addstats(cat, "snr") # We need this even without select, to show the snr_mean axis

#print momentsml.tools.table.info(cat)


if select:
	
	s = momentsml.tools.table.Selector("snr_mean > 10", [
		("min", "snr_mean", 10.0),
	])
	cat = s.select(cat)


if select:
	snr_mean = Feature("snr_mean", 5, 200, nicename=r"$\langle$S/N$\rangle$")
else:
	snr_mean = Feature("snr_mean", 0, 200, nicename=r"$\langle$S/N$\rangle$")
tru_flux = Feature("tru_flux", nicename=r"$F$ [counts]")
tru_rad = Feature("tru_rad", 0, 11, nicename=r"Half-light radius $R$ [pix]")
#tru_radwrtPSF = Feature("tru_radwrtPSF", 0.5, 3.5, nicename=r"$R/R_\mathrm{PSF}$")
tru_sersicn = Feature("tru_sersicn", 0, 7, nicename=r"S\'ersic index $n$")
tru_g = Feature("tru_g", -0.02, 0.72, nicename=r"Galaxy ellipticity $|\varepsilon|$")


cat["tru_radwrtPSF"] = cat["tru_rad"] / (cat["tru_psf_sigma"] * 1.1774)

tru_s1 = Feature("tru_s1")
tru_s2 = Feature("tru_s2")
pre_s1 = Feature("pre_s1", rea="all")
pre_s2 = Feature("pre_s2", rea="all")


if useweights:
	pre_s1w = Feature("pre_s1w", rea="all")
	pre_s2w = Feature("pre_s2w", rea="all")
else:
	pre_s1w = None
	pre_s2w = None
	
	

def make_plot(ax, featbin, showlegend=False, sersic=False):
	ax.axhline(0.0, color='gray', lw=0.5)	
	if sersic: # We don't let the code make bins, but provide binlims to separate the discrete cases:
		binlims = np.linspace(0.29, 6.1, 5)
		showbins = False
	else:
		showbins = True
		binlims = None
	# The first component
	momentsml.plot.mcbin.mcbin(ax, cat, tru_s1, pre_s1, featbin, featprew=pre_s1w, comp=1, binlims=binlims, showbins=showbins, regressmethod=regressmethod)
	# For the second component, we do not overplot bins again, but we might show the legend
	momentsml.plot.mcbin.mcbin(ax, cat, tru_s2, pre_s2, featbin, featprew=pre_s2w, comp=2, binlims=binlims, showbins=False, showlegend=showlegend, regressmethod=regressmethod)
	momentsml.plot.mcbin.make_symlog(ax, featbin)
	ax.set_xlabel(featbin.nicename)


fig = plt.figure(figsize=(8, 6))
plt.subplots_adjust(
	left  = 0.12,  # the left side of the subplots of the figure
	right = 0.99,    # the right side of the subplots of the figure
	bottom = 0.1,   # the bottom of the subplots of the figure
	top = 0.95,      # the top of the subplots of the figure
	wspace = 0.08,   # the amount of width reserved for blank space between subplots,
	                # expressed as a fraction of the average axis width
	hspace = 0.25,   # the amount of height reserved for white space between subplots,
					# expressed as a fraction of the average axis heightbottom=0.1, right=0.8, top=0.9)
	)

ax = plt.subplot(2, 2, 1)
make_plot(ax, snr_mean, showlegend=True)
ax.set_ylabel("Metric value")

ax = plt.subplot(2, 2, 2)
make_plot(ax, tru_rad)
ax.set_yticklabels([])

ax = plt.subplot(2, 2, 3)
make_plot(ax, tru_sersicn, sersic=True)
ax.set_ylabel("Metric value")

ax = plt.subplot(2, 2, 4)
make_plot(ax, tru_g)
ax.set_yticklabels([])


momentsml.plot.figures.savefig(os.path.join(config.valdir, valname + "_cond_biases"), fig, fancy=True)
plt.show()











"""
for comp in ["1","2"]:

	# If no weights are in the catalog (or not yet), we add ones
	if not "pre_s{}w".format(comp) in cat.colnames:
		
		# First putting all weights to 1.0:
		cat["pre_s{}w".format(comp)] = np.ones(cat["adamom_g1"].shape)
		
	cat["pre_s{}w_norm".format(comp)] = cat["pre_s{}w".format(comp)] / np.max(cat["pre_s{}w".format(comp)])

	momentsml.tools.table.addrmsd(cat, "pre_s{}".format(comp), "tru_s{}".format(comp))
	momentsml.tools.table.addstats(cat, "pre_s{}".format(comp), "pre_s{}w".format(comp))
	cat["pre_s{}_wbias".format(comp)] = cat["pre_s{}_wmean".format(comp)] - cat["tru_s{}".format(comp)]
"""

"""
#for comp in ["1", "2"]:
#	pass
	#cat["pre_g{}".format(comp)] = cat["pre_g{}_adamom".format(comp)]
	#momentsml.tools.table.addstats(cat, "pre_g{}".format(comp))
	#momentsml.tools.table.addrmsd(cat, "pre_g{}".format(comp), "tru_s{}".format(comp))
"""


"""
s = momentsml.tools.table.Selector("ok", [
	("min", "snr_mean", 10),
	#("in", "tru_rad", 0, 10.),
	#("max", "adamom_frac", 0.005)
	]
)
cat = s.select(cat)
"""

"""
#--------------------------------------------------------------------------------------------------
fig = plt.figure(figsize=(8.5, 3))
plt.subplots_adjust(wspace=0.0)
plt.subplots_adjust(bottom=0.2)
plt.subplots_adjust(right=0.92)
plt.subplots_adjust(left=0.11)

#------------------------------------------------------------
maxy = 0.015#cat["pre_g{}_bias".format(component)].max() * 1.06
miny = -0.015#cat["pre_g{}_bias".format(component)].min() * 1.02

maxsnr = cat["snr_mean"].max()
minsnr = cat["snr_mean"].min()

minshear = -0.12
maxshear = 0.12
#------------------------------------------------------------

ax = fig.add_subplot(1, 2, 1)
ax.fill_between([-1, 1], -2e-3, 2e-3, alpha=0.2, facecolor='darkgrey')
ax.axhline(0, ls='--', color='k')
momentsml.plot.scatter.scatter(ax, cat, main_feat,  Feature("pre_s{}_bias".format(component)), \
	featc=Feature("snr_mean"), marker='.', cmap="plasma", hidecbar=True, vmin=minsnr, vmax=maxsnr)
ax.set_xlabel(r"True shear $g_{%s}$" % component)
ax.set_ylabel(r"Shear bias")


metrics = momentsml.tools.metrics.metrics(cat, main_feat,  Feature("pre_s{}_bias".format(component)), pre_is_res=True)

ax.annotate(r"$\mathrm{RMSD=%.5f}$" % metrics["rmsd"], xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -4), textcoords='offset points', ha='left', va='top')
ax.annotate(r"$10^3\mu=%.1f \pm %.1f;\,10^3c=%.1f \pm %.1f$" % (metrics["m"]*1000.0, metrics["merr"]*1000.0, \
	metrics["c"]*1000.0, metrics["cerr"]*1000.0), xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -19), textcoords='offset points', ha='left', va='top')
#ax.annotate(r"$10^3c=%.1f \pm %.1f$" % (metrics["c"]*1000.0, metrics["cerr"]*1000.0), xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -35), textcoords='offset points', ha='left', va='top')
ax.set_ylim([miny, maxy])
ax.set_xlim([minshear, maxshear])

#------------------------------------------------------------

ax = fig.add_subplot(1, 2, 2)
ax.fill_between([-1, 1], -2e-3, 2e-3, alpha=0.2, facecolor='darkgrey')
ax.axhline(0, ls='--', color='k')
momentsml.plot.scatter.scatter(ax, cat, main_feat,  Feature("pre_s{}_bias".format(component2)), 
	featc=Feature("snr_mean", nicename=r"S/N"), marker='.', cmap="plasma", vmin=minsnr, vmax=maxsnr)
ax.set_xlabel(r"True shear $g_{%s}$" % component2)
metrics = momentsml.tools.metrics.metrics(cat, main_feat,  Feature("pre_s{}_bias".format(component2)), pre_is_res=True)

ax.annotate(r"$\mathrm{RMSD=%.5f}$" % metrics["rmsd"], xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -4), textcoords='offset points', ha='left', va='top')
ax.annotate(r"$10^3\mu=%.1f \pm %.1f;\,10^3c=%.1f \pm %.1f$" % (metrics["m"]*1000.0, metrics["merr"]*1000.0, \
	metrics["c"]*1000.0, metrics["cerr"]*1000.0), xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -19), textcoords='offset points', ha='left', va='top')
ax.set_ylim([miny, maxy])
ax.set_xlim([minshear, maxshear])
ax.set_yticklabels([])
ax.set_ylabel("")

momentsml.plot.figures.savefig(os.path.join(config.valdir, valname + "_plot_6a"), fig, fancy=True, pdf_transparence=True)
#------------------------------------------------------------
"""

"""




#------------------------------------------------------------
isubfig = 1
ncol = 2
nlines = int(np.ceil(len(param_feats) / (ncol*1.)))
fig = plt.figure(figsize=(4.1*ncol, 3 * nlines))
plt.subplots_adjust(wspace=0.07)
plt.subplots_adjust(hspace=0.27)
plt.subplots_adjust(right=0.98)
plt.subplots_adjust(top=0.98)
import matplotlib.transforms as mtransforms
no_legend = True

assert "res" not in cat.colnames
lintresh = 2e-3

coln = 0
for iplot, featc in enumerate(param_feats):
	
	ax = fig.add_subplot(nlines, ncol, isubfig + iplot)
	
	trans = mtransforms.blended_transform_factory(ax.transAxes, ax.transData)
	ax.fill_between([0, 1], -lintresh, lintresh, alpha=0.2, facecolor='darkgrey', transform=trans)
	ax.axhline(0, ls='--', color='k')
	ax.set_ylabel(r"Shear bias")
	
	for icomp, comp in enumerate(["1", "2"]):
		
		main_pred = "s{}".format(comp)
		maincol = "tru_{}".format(main_pred)
		main_feat = Feature(maincol)
	
		cat["res"] = cat["pre_s{}_bias".format(comp)]
		
		#xbinrange = momentsml.plot.utils.getrange(cat, featc)
		#binsumma = momentsml.plot.utils.summabin(cat[maincol], cat["res"], xbinrange=xbinrange, nbins=nbins)
		
		cbinrange = momentsml.plot.utils.getrange(cat, featc)
		
		if featc.colname is "tru_sersicn":
			cbinlims = np.linspace(1-0.1, 4+0.1, 11)	
		else:
			cbinlims = np.array([np.percentile(cat[featc.colname], q) for q in np.linspace(0.0, 100.0, ncbins+1)])
	
		#print cbinlims
		
		cbinlows = cbinlims[0:-1]
		cbinhighs = cbinlims[1:]
		#cbincenters = 0.5 * (cbinlows + cbinhighs)
		#assert len(cbincenters) == ncbins
		
		#print cbincenters
		
		
		#offsetscale = 0.5*((xbinrange[1] - xbinrange[0])/float(nbins))/float(ncbins)
		
		cbinpointcenters = []
		ms = []
		merrs = []
		cs = []
		cerrs = []
		for i in range(ncbins):
			
			#offset = (i - float(ncbins)/2) * offsetscale
			
			# We build the subset of data that is in this color bin:
			selcbin = momentsml.tools.table.Selector(featc.colname, [("in", featc.colname, cbinlows[i], cbinhighs[i])])
			cbindata = selcbin.select(cat)
			
			#if len(cbindata) == 0:
			#	continue
			# WTF dangerous!
			
			cbinfrac = float(len(cbindata)) / float(len(cat))
			
			# And we perform the linear regression
			md = momentsml.tools.metrics.metrics(cbindata,
				main_feat, # Redefining those to get rid of any rea settings that don't apply to cbindata
				momentsml.tools.feature.Feature("res"),
				pre_is_res=True)
			
			ms.append(md["m"])
			merrs.append(md["merr"])
			cs.append(md["c"])
			cerrs.append(md["cerr"])
			
			cbinpointcenters.append(np.mean(cbindata[featc.colname]))
			
			#ax.plot(np.array(xbinrange), md["m"]*np.array(xbinrange)+md["c"], color=color, ls="-")
			
			
			
			# And now bin this in x:
			#cbinsumma = momentsml.plot.utils.summabin(cbindata[maincol], cbindata["res"], xbinrange=xbinrange, nbins=nbins, equalcount=False)
			
		if icomp == 0:
			markm = '*'
			markc = '^'
			color = 'k'
		elif icomp == 1:
			markm = 's'
			markc = 'v'
			color = 'r'
		labelc = r"$c_{%s}$" % (comp)
		labelm = r"$\mu_{%s}$" % (comp)
			
		ax.errorbar(cbinpointcenters, ms, yerr=merrs, color=color, marker=markm, label=labelm)
		ax.errorbar(cbinpointcenters, cs, yerr=cerrs, color=color, marker=markc, ls=':', label=labelc)


	# Outside of the "comp" loop, we add bins
	for x in cbinlims:
		ax.axvline(x, color='gray', lw=0.5)

	#if (isubfig + iplot)/2+1 == nlines:
	#	ax.set_xlabel(r"$\mathrm{True\ shear}$")
	#else:
	#	ax.set_xticklabels([])
	
	
	ax.set_yscale('symlog', linthreshy=lintresh)
	ax.set_xlabel(featc.nicename)
	if featc.colname == "tru_g" or (iplot == len(param_feats) + 1 and no_legend):
		plt.legend(loc="best", handletextpad=0.07,fontsize="small", framealpha=0.5, columnspacing=0.1, ncol=2)
		no_legend = False
	
	ax.set_ylim([-1e-1, 1e-1])
	
	ticks = np.concatenate([np.arange(-lintresh, lintresh, 1e-3)])#, np.arange(lintresh, 1e-2, 9)])
	s = ax.yaxis._scale
	ax.yaxis.set_minor_locator(ticker.SymmetricalLogLocator(s, subs=[1., 2.,3.,4.,5.,6.,7.,8.,9.,-2.,-3.,-4.,-5.,-6.,-7.,-8.,-9.]))
	ticks = np.concatenate([ticks, ax.yaxis.get_minor_locator().tick_values(-.1, .1)])
	ax.yaxis.set_minor_locator(ticker.FixedLocator(ticks))
	
	if featc.low is not None and featc.high is not None:
		ax.set_xlim([featc.low, featc.high])
		#ax.locator_params(axis='x', nticks=2)
		#tick_spacing = (featc.high - featc.low) / 5.
		#ax.xaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))
		#ax.xaxis.set_major_formatter(ticker.FormatStrFormatter(r'$%0.3f$'))
		#ax.xaxis.set_major_formatter(ticker.FormatStrFormatter(r'$%0.3f$'))
		
	if coln > 0:
		ax.set_ylabel("")
		ax.set_yticklabels([])

	coln += 1
	if coln == ncol: coln = 0


momentsml.plot.figures.savefig(os.path.join(config.valdir, valname + "_condbias_b"), fig, fancy=True, pdf_transparence=True)
plt.show()
"""



"""

WNet is deprecated, should be updated and tested if required.

A WNet is a custom network (in fact 2 networks in parallel), that also predicts a weight for each output parameter.

"""

import numpy as np
import scipy.optimize
from datetime import datetime

import logging
logger = logging.getLogger(__name__)

from . import layer
from . import net
from . import utils
from . import err
from . import act
from . import data


class WNet():
	"""
	A WNet contains 2 Nets.
	To a training, it looks like a normal Net but with twice the number of ouputs (as it returns outputs and weights).
	The appropriate error fuctions know how to combine these into a single metric.
	"""
	
	def __init__(self, ni, nhs, no=1, name=None, inames=None, onames=None, netokwargs={}, netwkwargs={}):
		"""
		:param ni: Number of input features
		:param nhs: Numbers of neurons in hidden layers
		:type nhs: tuple
		:param no: Number of ouput neurons *NOT* taking into account the weights.
			
		:param name: if None, will be set automatically
		:type name: string
		
		:param inames: a list of names (strings) for the input nodes, to be used, e.g., in checkplots.
			These names have a purely decorative function, and are optional.
		:param onames: idem, for the ouptut nodes.
		
		:param netokwargs: dict of further kwargs for the constructor of the Net yielding the outputs
		:param netwkwargs: dict of further kwargs for the constructor of the Net yielding the weights
		
		
		"""
	
		self.ni = ni
		self.no = 2*no	# For every output we also have a weight. Assuming this is mandatory for the
						# way we return outputs, and so we'll stick to this and not make it more complicated.
						# However we do not rely on the fact the that two Nets are "similar"!
						# So it might be that the two nets have different numbers of params!
	
		logger.info("Building a WNet...")
		self.neto = net.Net(ni, nhs, no=no, name="neto", inames=inames, onames=onames, **netokwargs)
		self.netw = net.Net(ni, nhs, no=no, name="netw", inames=inames, onames=onames, **netwkwargs)
	
		self.name = name
		
		# We adapt the onames of the 2 networks:
		
		for i in range(self.neto.no):
			self.neto.onames[i] += "_o"
			self.netw.onames[i] += "_w"
			
	
	def __str__(self):
		"""
		A short string describing the network
		"""
		
		autotxt = "WNet" + str(self.neto)[7:]
			
		if self.name is None:
			return autotxt
		else:
			return "'{name}' {autotxt}".format(name=self.name, autotxt=autotxt)

	
	def report(self):
		"""
		Returns a text about the network parameters, useful for debugging.
		"""
		txt = ["#"*120, self.neto.report(), self.netw.report(), "#"*120]
		return "\n".join(txt)
		
	
	def save(self, filepath):
		"""
		Saves self into a pkl file
		"""
		utils.writepickle(self, filepath)		
	
	
	def nparams(self):
		"""
		Returns the number of parameters of the network
		"""
		return self.neto.nparams() + self.netw.nparams() 
		
	
	def get_params_ref(self, mode=None):
		"""
		Get a single 1D numpy array containing references to all network weights and biases.
		Note that each time you call this, you loose the "connection" to the ref from any previous calls.
		
		:param mode: switch to get only the params for the predictive outputs ("o"), or only the params for the weights ("w"), or both (None)
			
		For mode "None", I did not manage to get this working by reusing the equivalent functions of the two Net objects.
		So redoing from scratch. We put the "o" params first, then the "w" params.
		
		"""
		
		if mode == "o":
			return self.neto.get_params_ref()
		elif mode == "w":
			return self.netw.get_params_ref()
		elif mode == None:
		
			ref = np.empty(self.nparams())
			ind = 0
		
			# We start with the "o" params:
			for l in self.neto.layers[::-1]:
			
				ref[ind:ind+l.nn] = l.biases.flatten() # makes a copy
				ref[ind+l.nn:ind+l.nparams()] = l.weights.flatten() # makes a copy
				l.biases = ref[ind:ind+l.nn] # a view
				l.weights = ref[ind+l.nn:ind+l.nparams()].reshape(l.nn, l.ni) # a view
				ind += l.nparams()
			
			# And now the "w" params:
			assert ind == self.neto.nparams()
			
			for l in self.netw.layers[::-1]:
				
				ref[ind:ind+l.nn] = l.biases.flatten() # makes a copy
				ref[ind+l.nn:ind+l.nparams()] = l.weights.flatten() # makes a copy
				l.biases = ref[ind:ind+l.nn] # a view
				l.weights = ref[ind+l.nn:ind+l.nparams()].reshape(l.nn, l.ni) # a view
				ind += l.nparams()
		
			assert ind == self.nparams()
			return ref
		
		else:
			raise ValueError("Unknown get_params_ref mode !")


	def get_paramlabels(self):
		"""
		Returns a list with labels describing the params.
		
		"""
		paramlabels = []
		paramlabels.extend(["neto_" + l for l in self.neto.get_paramlabels()])
		paramlabels.extend(["netw_" + l for l in self.netw.get_paramlabels()])
		return paramlabels
		
	def get_paramslice(self, mode=None):
		"""
		Would need to be implemented !
		Code below is just copy an paste from an old train method.
		
		"""
		raise RuntimeError("Implement me!")
		
		
#		logger.info("Preparing a paramslice to mode '{}'...".format(mode))
#		
#		if mode == "o": # the slice selects only the params of the "ouputs"
#			self.paramslice = slice(0, self.net.neto.nparams())
#		elif mode == "w": # Idem but for the weights
#			self.paramslice = slice(self.net.neto.nparams(), self.net.nparams())
#		elif mode == None: # Empty slice, use all params
#			self.paramslice = slice(None)
#		else:
#			raise ValueError("Unknown mode!")
#		
#		logger.info("Set paramslice to mode '{}' : {}/{} params are free to be optimized.".format(
#			mode, len(self.params[self.paramslice]), self.net.nparams())
#			)
		

		
	def addnoise(self, **kwargs):
		"""
		Adds random noise to all parameters.
		"""
	
		self.neto.addnoise(**kwargs)
		self.netw.addnoise(**kwargs)
			
			
	def setini(self):
		"""
		Adjusts the network parameters so to approximatively get:
		- identity for neto
		- zero for netw (i.e., always return 0.0 -> a weight of e0 = 1, no matter what the input is).
		
		"""

		self.neto.setidentity()
		for l in self.netw.layers:
			l.zero()
			

	def run(self, inputs):
		"""
		Similar to Net.run().
		We compute outputs for both Nets, and combine these into a single array, mimicing a single Net with twice the number of outputs.
		So for the outputs, along the second dimension (nodes), we first have the normal outputs, and then the associated weights.
		"""
		
		if inputs.ndim != 3:
			raise ValueError("Sorry, for WNet I want 3D inputs only.")
		
		os = self.neto.run(inputs)
		ws = self.netw.run(inputs)
		
		assert os.shape == (inputs.shape[0], self.neto.no, inputs.shape[2])
		assert ws.shape == (inputs.shape[0], self.netw.no, inputs.shape[2])
		
		# And we stack the two parts 
		outputs = np.concatenate((os, ws), axis=1)
		
		assert outputs.shape == (inputs.shape[0], self.neto.no + self.netw.no, inputs.shape[2])
		
		return outputs
		
			
	
	def predict(self, inputs):
		"""
		We compute the outputs from the inputs using self.run, but here we do take care of the potential mask.
		
		This is never used during the training phase.
		
		:param inputs: a (potentially masked) 3D array
		
		:returns: a 3D array, appropriatedly masked
		
		"""
		
		# To be adapted for WNet...
		"""
		logger.info("Predicting with input = {intype} of shape {inshape}".format(
			intype=str(type(inputs)), inshape=str(inputs.shape)))

		if inputs.ndim != 3:
			raise ValueError("Sorry, I only accept 3D input")

		(inputs, outputsmask) = data.demask(inputs, no=self.no)
		
		# We can simply run the network with the unmasked inputs:
		
		logger.info("Running the actual predictions...")
		outputs = self.run(inputs)
		
		# And now mask these outputs, if required:
		
		if outputsmask is not None:
			outputs = np.ma.array(outputs, mask=outputsmask)
		
		return outputs
		"""



"""
Simulates a dedicated training or validation set for each subfield.
This scirpt takes an argument, run it with -h to get help.
"""

import argparse

import momentsml
import momentsml.meas
import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.INFO)
logger = logging.getLogger(__name__)

import simparams
import measfcts



def run(simtype=None):
	
	simname = config.datasets[simtype]
	
	logger.info("Drawing simtype {} => {}...".format(simtype, simname))
	
	
	############## First, the fiducial sets, as small and fast as possible, to give a baseline.
	
	if simname == "ts-ell-nn-train-rea10":
		
		# The fiducial train-shear: **** ellipticity training, without noise **** (yes, that's a big compromise, for GREAT3).
		# The "nn" stands for "no noise"
		sp = simparams.G3Sersics(
			name = simname,
			snc_type = 1,
			shear = 0,
			noise_level = 0,
			obstype = config.great3.obstype,
			distmode = "train"
		)
		# 1000 cases with 10 rea
		drawconf = {
			"n":1000,
			"nc":10,
			"nrea":10,
			"ncat":1,
			"ncpu":10,
			"groupmode":"g"			
		}

	elif simname == "vs-shear-n-G3-snc1000":
	
		# The fiducial valid-shear. Heavy !
		sp = simparams.G3Sersics(
			name = simname,
			snc_type = 1000, # A lot, but this also covers noise and sub-pixel alignment. Hard to get more.
			shear = 0.1,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 1000 cases (different galaxies and shears), 1000 snc realizations
		drawconf = {
			"n": 20, # Don't put this to one, otherwise only one sersicn is used...
			"nc":1,
			"nrea":1,
			"ncat":50,
			"ncpu":25,
			"groupmode":"s"				
		}
	
	
	elif simname == "tw-200c-1000r":
	
		# The ficducial train-weight. Heavy, but can't make this smaller. Cases have different shear, but mix galaxies
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 4,
			shear = 0.1,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 200 cases, 1000 realizations with small snc
		drawconf = {
			"n":250,
			"nc":10,
			"nrea":1,
			"ncat":200,
			"ncpu":10,
			"groupmode":"s"				
		}

	
	elif simname == "vo-200c-8000r":

		# For overall validation, we almost mimick a full GREAT3 branch, huge!
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 4,
			shear = 0.1,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 200 cases, 8000 realizations.
		drawconf = {
			"n":2000,
			"nc":50,
			"nrea":1,
			"ncat":200,
			"ncpu":25,
			"groupmode":"s"				
		}


	elif simname == "simobscompa-G3":

		# Drawing 10'000 galaxies to compare their feature distribs with those of the "observations". Just for plots and tests.
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 1,
			shear = 0.0, # Would be OK to put a little shear, in fact, as the observations have some.
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 10'000 galaxies, single realization
		drawconf = {
			"n":1000,
			"nc":5,  # To get a good samplign of sersicn (n / nc)
			"nrea":1,
			"ncat":10,
			"ncpu":10,
			"groupmode":None
		}

	elif simname == "simobscompa-train":
		
		# As above, but with "train" distributions (instead of G3). Just for plots and tests.
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 1,
			shear = 0.0, # Would be OK to put a little shear, in fact, as the observations have some.
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "train"
		)
		# 10'000 galaxies, single realization
		drawconf = {
			"n":1000,
			"nc":5,  # To get a good samplign of sersicn (n / nc)
			"nrea":1,
			"ncat":10,
			"ncpu":10,
			"groupmode":None
		}

	########### Mimicing GREAT3 subfields

	elif simname == "sersicG3subfield":
		
		# We mimic a GREAT3 subfield. Single static shear, 10'000 gals, with SNC 2
		# Run this for every "subfield".
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 2,
			shear = 0.05,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 10'000 galaxies, single realization
		drawconf = {
			"n":5000, # 2x, with SNC
			"nc":50,  # To get a good samplign of sersicn (n / nc)
			"nrea":1,
			"ncat":1, # To get a single shear value
			"ncpu":1,
			"groupmode":None
		}

	elif simname == "sersicG3subfield_nosnc":
		
		# Idem, but without SNC, to check for differences
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 1,
			shear = 0.05,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 10'000 galaxies, single realization
		drawconf = {
			"n":10000,
			"nc":100,  # To get a good samplign of sersicn (n / nc)
			"nrea":1,
			"ncat":1, # To get a single shear value
			"ncpu":1,
			"groupmode":None
		}
	
	########## Now, some alternative datasets, to test out other approaches.
	
	elif simname == "ts-shear-nn-train-snc20":
		
		# For train-shear
		# Really train for shear, not for ellipticity. Still without noise here.
		# A bit close to what you ideally want to do.
		# Cases have different shear and different galaxies (one per case). Train distribs.
		sp = simparams.G3Sersics(
			name = simname,
			snc_type = 20,
			shear = 0.1,
			noise_level = 0,
			obstype = config.great3.obstype,
			distmode = "train"
		)
		# 1000 cases with 20 snc rea
		drawconf = {
			"n":40, # Don't put this to zero, otherwise only one sersicn is used.
			"nc":1,
			"nrea":1,
			"ncat":25,
			"ncpu":25,
			"groupmode":"s"			
		}

	
	elif simname == "ts-ell-n-train-rea100":		
		
		# Something we can afford, with noise, to see if noise-bias gets reduced: 
		sp = simparams.G3Sersics(
			name = simname,
			snc_type = 1,
			shear = 0,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "train"
		)
		# 1000 cases with 100 rea
		drawconf = {
			"n":1000,
			"nc":10,
			"nrea":100,
			"ncat":1,
			"ncpu":20,
			"groupmode":"g"			
		}
	
	
	elif simname == "tw-100c-10000r":
		
		# A huge weight training set
		# Cases have different shear, but mix galaxies
		# Experimenting with much more realizations, to see if bias on vo persists...
		# Seems to work, but expensive, and overall improvement is not interesting for GREAT3.
		
		sp = simparams.G3Sersics_statshear(
			name = simname,
			snc_type = 4,
			shear = 0.1,
			noise_level = 1,
			obstype = config.great3.obstype,
			distmode = "G3"
		)
		# 100 cases, 10000 realizations with small snc
		drawconf = {
			"n":2500,
			"nc":50,
			"nrea":1,
			"ncat":100,
			"ncpu":10,
			"groupmode":"s"				
		}
		
		
	else:
		raise RuntimeError("Unknown simname, see code for defined names.")

	for subfield in config.great3.subfields:
		runsub(subfield, sp, drawconf)


def runsub(subfield, sp, drawconf):
	
	
	# We have to read in the obs catalog of this subfield to get the noise of the sky:
	if sp.noise_level != 0:
		obscat = momentsml.tools.io.readpickle(config.great3.subpath(subfield, "obs", "img_meascat.pkl"))
		sp.noise_level = np.ma.mean(obscat["skymad"])
		logger.info("Noise level set to {}".format(sp.noise_level))
	
	# Getting the path to the correct directories
	simdir = config.great3.subpath(subfield, "sim")
	measdir = config.great3.subpath(subfield, "simmeas")
	
	# Loading the PSF for the subfield
	psfcat = momentsml.tools.io.readpickle(config.great3.subpath(subfield, "obs", "star_meascat.pkl"))
	
	
	# Simulating images
	momentsml.sim.run.multi(
		simdir=simdir,
		simparams=sp,
		drawcatkwargs={"n":drawconf["n"], "nc":drawconf["nc"], "stampsize":config.great3.stampsize()},
		drawimgkwargs={}, 
		psfcat=psfcat, psfselect="random",
		ncat=drawconf["ncat"], nrea=drawconf["nrea"], ncpu=drawconf["ncpu"],
		savepsfimg=False, savetrugalimg=False
	)


	# Measuring the newly drawn images
	momentsml.meas.run.onsims(
		simdir=simdir,
		simparams=sp,
		measdir=measdir,
		measfct=measfcts.gal,
		measfctkwargs={"branch":config.great3},
		ncpu=drawconf["ncpu"],
		skipdone=config.great3.skipdone
	)
	

	cat = momentsml.meas.avg.onsims(
		measdir=measdir, 
		simparams=sp,
		task="group",
		groupcols=measfcts.groupcols, 
		removecols=measfcts.removecols
	)

	momentsml.tools.table.keepunique(cat)
	momentsml.tools.io.writepickle(cat, os.path.join(measdir, sp.name, "groupmeascat.pkl"))
	
	#cat = momentsml.tools.io.readpickle(os.path.join(measdir, sp.name, "groupmeascat.pkl"))
	#print momentsml.tools.table.info(cat)
	#exit()

	# For shear sims, we group the results in cases of same true shears
	if drawconf["groupmode"] == "s":
			
		cat = momentsml.tools.table.groupreshape(cat, groupcolnames=["tru_s1", "tru_s2"])
		momentsml.tools.table.keepunique(cat)
		
	if drawconf["groupmode"] in ["s", "g"]:
	
		# For each case, we add the fraction of failed measurements
		nrea = cat["adamom_g1"].shape[1]
		logger.info("We have {} realizations".format(nrea))
		cat["adamom_failfrac"] = np.sum(cat["adamom_g1"].mask, axis=1) / float(nrea)
		#print cat["adamom_failfrac"]

		#print momentsml.tools.table.info(cat)
		momentsml.tools.io.writepickle(cat, os.path.join(measdir, sp.name, "groupmeascat.pkl")) # Just overwrite, don't need the other one
	
	"""
	# For statell sims, we group by ellipticity
	if "statell" in sp.name:
		
		cat = momentsml.tools.table.groupreshape(cat, groupcolnames=["tru_g1", "tru_g2"])
		momentsml.tools.table.keepunique(cat)
		#print momentsml.tools.table.info(cat)
		momentsml.tools.io.writepickle(cat, os.path.join(measdir, sp.name, "groupmeascat.pkl")) # Just overwrite, don't need the other one
	"""
	


if __name__ == '__main__':

	possible_simtypes = ["train-shear", "valid-shear", "train-weight", "valid-overall", "simobscompa", "mimic-great3"]

	parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	parser.add_argument('simtype', help='Type of dataset to simulate, must be in {}'.format(possible_simtypes))
	args = parser.parse_args()
	
	if args.simtype not in possible_simtypes:
		raise RuntimeError("Unknown simtype, must be in {}".format(possible_simtypes))
	
	run(simtype=args.simtype)




import os
import numpy as np
import matplotlib.ticker as ticker
import matplotlib.transforms as mtransforms

import momentsml.plot
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)


outdirplots = os.path.join('.', "plots")

momentsml.plot.figures.set_fancy(16)

metrics_12 = {
	'fGC':{'m1':-1.50, 'm1err':2.78, 'm2':3.18, 'm2err':2.04, 'c1':0.02, 'c1err':0.08, 'c2':0.01, 'c2err':0.06},
	'CGC':{'m1':4.34, 'm1err':3.52, 'm2':1.99, 'm2err':3.03, 'c1':0.02, 'c1err':0.09, 'c2':-0.11, 'c2err':0.08},
	'RGC':{'m1':-5.42, 'm1err':3.15, 'm2':-6.40, 'm2err':3.15, 'c1':0.12, 'c1err':0.09, 'c2':-0.11, 'c2err':0.08},
	'fSC':{'m1':-0.97, 'm1err':1.49, 'm2':-1.87, 'm2err':1.50, 'c1':-0.04, 'c1err':0.04, 'c2':-0.00, 'c2err':0.05},
	'CSC':{'m1':-3.11, 'm1err':3.55, 'm2':-3.73, 'm2err':3.78, 'c1':-0.19, 'c1err':0.09, 'c2':-0.00, 'c2err':0.09},
	'RSC':{'m1':-5.49, 'm1err':3.15, 'm2':-6.40, 'm2err':3.15, 'c1':0.04, 'c1err':0.08, 'c2':0.04, 'c2err':0.08},
	}
																		
metrics_pluscross = { 
	'fGC':{'m1':1.97, 'm1err':2.28, 'm2':-0.30, 'm2err':2.62, 'c1':0.08, 'c1err':0.07, 'c2':-0.02, 'c2err':0.07},
	'CGC':{'m1':1.58, 'm1err':3.43, 'm2':3.70, 'm2err':3.06, 'c1':-0.03, 'c1err':0.09, 'c2':-0.16, 'c2err':0.08},
	'RGC':{'m1':-3.84, 'm1err':3.34, 'm2':-7.98, 'm2err':3.79, 'c1':0.29, 'c1err':0.08, 'c2':-0.02, 'c2err':0.09},
	'fSC':{'m1':-1.55, 'm1err':1.41, 'm2':-1.18, 'm2err':1.57, 'c1':0.07, 'c1err':0.04, 'c2':-0.02, 'c2err':0.04},
	'CSC':{'m1':-0.04, 'm1err':3.73, 'm2':-5.37, 'm2err':3.52, 'c1':-0.35, 'c1err':0.09, 'c2':-0.09, 'c2err':0.09},
	'RSC':{'m1':-3.77, 'm1err':3.08, 'm2':-7.70, 'm2err':3.25, 'c1':-0.01, 'c1err':0.08, 'c2':-0.08, 'c2err':0.08},
	}


metrics_mean = {}
for key in metrics_12.keys():
	metrics_mean[key] = {}
	metrics_mean[key]['m1'] = (metrics_12[key]['m1'] + metrics_12[key]['m2'])/2
	metrics_mean[key]['m1err'] = (metrics_12[key]['m1err'] + metrics_12[key]['m2err'])/2
	metrics_mean[key]['c1'] = metrics_pluscross[key]['c1']
	metrics_mean[key]['c1err'] = metrics_pluscross[key]['c1err']
						
print metrics_mean
if not os.path.exists(outdirplots):
	os.mkdir(outdirplots)
	

order = ['fGC', 'CGC', 'RGC', 'fSC', 'CSC', 'RSC']
metrics = [metrics_12, metrics_pluscross, metrics_mean]
outname = ["12", "pc", "mmean"]
xlabels = [{1:'c_1', 2:'c_2'}, {1:'c_+', 2:'c_\\times'}, {1:"c_+"}]
ylabels = [{1:'\\mu_1', 2:'\\mu_2'}, {1:'\\mu_+', 2:'\\mu_\\times'}, {1:"\\langle\\mu\\rangle"}]
colours = ['k', 'orange', 'limegreen', 'turquoise', "royalblue", 'r', "#a65628", "#f781bf", "#4daf4a"]
colours = ['maroon', 'navy', 'darkgreen', 'tomato', "steelblue", 'lime', "#a65628", "#f781bf", "#4daf4a"]
markers = ['o','^',"D"] * 2#,'o','*','D']
ls = ['-','-',"-", '--', '--', '--']#,'o','*','D']

lintreshy = 2e-3
lintreshx = 2e-4

assert len(order) <= len(markers)
assert len(order) <= len(colours)
assert len(metrics) <= len(outname)
assert len(metrics) <= len(xlabels)
assert len(metrics) <= len(ylabels)

for imet, metric in enumerate(metrics):
	
	fig = plt.figure(figsize=(7*len(xlabels[imet]), 4.5))
	plt.subplots_adjust(wspace=0.2)
	plt.subplots_adjust(hspace=0.4)
	plt.subplots_adjust(right=0.98)
	plt.subplots_adjust(top=0.96)
	plt.subplots_adjust(bottom=0.12)
	plt.subplots_adjust(left=0.15)
	
	for comp in range(1, 1+len(xlabels[imet])):
	
		ax = fig.add_subplot(1, len(xlabels[imet]), comp)
		
		trans = mtransforms.blended_transform_factory(ax.transAxes, ax.transData)
		ax.fill_between([0, 1], -lintreshy, lintreshy, alpha=0.2, facecolor='darkgrey', transform=trans)
		ax.axvspan(-lintreshx, lintreshx, alpha=0.2, facecolor='darkgrey')
		ax.axhline(0, ls='--', color='k')
		ax.axvline(0, ls='--', color='k')
		
		ax.set_ylabel(r"Multiplicative bias ${%s}$" % ylabels[imet][comp])
		ax.set_xlabel(r"Additive bias ${%s}$" % xlabels[imet][comp])
	
		for ibranch, branch in enumerate(order):
		
		
			_,capsstuff,barlinecols = ax.errorbar(metric[branch]['c{}'.format(comp)] * 1e-3, metric[branch]['m{}'.format(comp)] * 1e-3, \
					xerr=metric[branch]['c{}err'.format(comp)] * 1e-3, yerr=metric[branch]['m{}err'.format(comp)] * 1e-3, \
					fmt=markers[ibranch], color=colours[ibranch], label=branch, elinewidth=1.5)
			
			barlinecols[0].set_linestyle(ls[ibranch])
			barlinecols[1].set_linestyle(ls[ibranch])
			for cap in capsstuff:
				cap.set_markeredgewidth(2)
			
	
		ax.set_yscale('symlog', linthreshy=lintreshy)
		ax.set_xscale('symlog', linthreshx=lintreshx)
	
		ax.set_ylim([-5e-1, 5e-1])
		ticks = np.concatenate([np.arange(-lintreshy, lintreshy, 1e-3)])#, np.arange(lintresh, 1e-2, 9)])
		s = ax.yaxis._scale
		ax.yaxis.set_minor_locator(ticker.SymmetricalLogLocator(s, subs=[1., 2.,3.,4.,5.,6.,7.,8.,9.,-2.,-3.,-4.,-5.,-6.,-7.,-8.,-9.]))
		ticks = np.concatenate([ticks, ax.yaxis.get_minor_locator().tick_values(-.1, .1)])
		ax.yaxis.set_minor_locator(ticker.FixedLocator(ticks))
		
		ax.set_xlim([-5e-1, 5e-1])
		ticks = np.concatenate([np.arange(-lintreshx, lintreshx, 1e-4)])#, np.arange(lintresh, 1e-2, 9)])
		s = ax.xaxis._scale
		ax.xaxis.set_minor_locator(ticker.SymmetricalLogLocator(s, subs=[1., 2.,3.,4.,5.,6.,7.,8.,9.,-2.,-3.,-4.,-5.,-6.,-7.,-8.,-9.]))
		ticks = np.concatenate([ticks, ax.xaxis.get_minor_locator().tick_values(-.1, .1)])
		ax.xaxis.set_minor_locator(ticker.FixedLocator(ticks))
		
		plt.legend(loc="best", handletextpad=0.15,fontsize="small", framealpha=0.5, columnspacing=0.1, ncol=2, numpoints=1)
		
	momentsml.plot.figures.savefig(os.path.join(outdirplots, "great3_cross_plot_{}".format(outname[imet])), fig, fancy=True, pdf_transparence=True)
plt.show()



"""
In/out functions
"""

import os
import cPickle as pickle
import astropy.io.fits
import gzip

import logging
logger = logging.getLogger(__name__)

def writepickle(obj, filepath, protocol = -1):
    """
    I write your python object obj into a pickle file at filepath.
    If filepath ends with .gz, I'll use gzip to compress the pickle.
    Leave protocol = -1 : I'll use the latest binary protocol of pickle.
    """
    if os.path.splitext(filepath)[1] == ".gz":
        pkl_file = gzip.open(filepath, 'wb')
    else:
        pkl_file = open(filepath, 'wb')
    
    pickle.dump(obj, pkl_file, protocol)
    pkl_file.close()
    logger.info("Wrote %s" % filepath)
    
def readpickle(filepath):
    """
    I read a pickle file and return whatever object it contains.
    If the filepath ends with .gz, I'll unzip the pickle file.
    """
    if os.path.splitext(filepath)[1] == ".gz":
        pkl_file = gzip.open(filepath,'rb')
    else:
        pkl_file = open(filepath, 'rb')
    obj = pickle.load(pkl_file)
    pkl_file.close()
    logger.info("Read %s" % filepath)
    return obj


def fromfits(filepath):
    """
    Read simple 1-hdu FITS files -> numpy arrays, so that the indexes [x,y] follow the orientations of
    x, y on ds9, respectively.
    """
    a = astropy.io.fits.getdata(filepath).transpose()
    logger.info("Read FITS images %s from file %s" % (a.shape, filepath))
    return a
    

def tofits(a, filepath):
    """
    Writes a simply 2D numpy array to FITS, same convention.
    """
    
    if os.path.exists(filepath):
        logger.warning("File %s exists, I will overwrite it!" % (filepath))

    astropy.io.fits.writeto(filepath, a.transpose(), clobber=1)
    logger.info("Wrote %s array to %s" % (a.shape, filepath))
